# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
The bigfish.stack subpackage includes functions to read data, preprocess them
and build stack of images.
"""

from .utils import check_array
from .utils import check_df
from .utils import check_parameter
from .utils import check_range_value
from .utils import check_recipe
from .utils import check_datamap
from .utils import get_margin_value
from .utils import get_eps_float32
from .utils import load_and_save_url
from .utils import check_hash
from .utils import compute_hash
from .utils import check_input_data
from .utils import moving_average
from .utils import centered_moving_average
from .utils import get_sigma
from .utils import get_radius

from .io import read_image
from .io import read_dv
from .io import read_array
from .io import read_uncompressed
from .io import read_cell_extracted
from .io import read_array_from_csv
from .io import read_dataframe_from_csv
from .io import save_image
from .io import save_array
from .io import save_cell_extracted
from .io import save_data_to_csv

from .preprocess import build_stacks
from .preprocess import build_stack
from .preprocess import build_stack_no_recipe
from .preprocess import rescale
from .preprocess import cast_img_uint8
from .preprocess import cast_img_uint16
from .preprocess import cast_img_float32
from .preprocess import cast_img_float64

from .filter import mean_filter
from .filter import median_filter
from .filter import maximum_filter
from .filter import minimum_filter
from .filter import log_filter
from .filter import gaussian_filter
from .filter import remove_background_mean
from .filter import remove_background_gaussian
from .filter import dilation_filter
from .filter import erosion_filter

from .projection import maximum_projection
from .projection import mean_projection
from .projection import median_projection
from .projection import focus_projection
from .projection import in_focus_selection
from .projection import get_in_focus_indices

from .illumination import compute_illumination_surface
from .illumination import correct_illumination_surface

from .postprocess import identify_objects_in_region
from .postprocess import remove_transcription_site
from .postprocess import extract_cell
from .postprocess import extract_spots_from_frame
from .postprocess import summarize_extraction_results
from .postprocess import center_mask_coord
from .postprocess import from_boundaries_to_surface
from .postprocess import from_surface_to_boundaries
from .postprocess import from_binary_to_coord
from .postprocess import complete_coord_boundaries
from .postprocess import from_coord_to_frame
from .postprocess import from_coord_to_surface

from .augmentation import augment_2d
from .augmentation import augment_2d_function
from .augmentation import augment_8_times
from .augmentation import augment_8_times_reversed

from .quality import compute_focus


_utils = [
    "check_array",
    "check_df",
    "check_recipe",
    "check_datamap",
    "check_parameter",
    "check_range_value",
    "get_margin_value",
    "get_eps_float32",
    "load_and_save_url",
    "check_hash",
    "compute_hash",
    "check_input_data",
    "moving_average",
    "centered_moving_average",
    "get_sigma",
    "get_radius"]

_io = [
    "read_image",
    "read_dv",
    "read_array",
    "read_uncompressed",
    "read_cell_extracted",
    "read_array_from_csv",
    "read_dataframe_from_csv",
    "save_image",
    "save_array",
    "save_cell_extracted",
    "save_data_to_csv"]

_preprocess = [
    "build_stacks",
    "build_stack",
    "build_stack_no_recipe",
    "rescale",
    "cast_img_uint8",
    "cast_img_uint16",
    "cast_img_float32",
    "cast_img_float64"]

_filter = [
    "log_filter",
    "mean_filter",
    "median_filter",
    "maximum_filter",
    "minimum_filter",
    "gaussian_filter",
    "remove_background_mean",
    "remove_background_gaussian",
    "dilation_filter",
    "erosion_filter"]

_projection = [
    "maximum_projection",
    "mean_projection",
    "median_projection",
    "in_focus_selection",
    "get_in_focus_indices",
    "focus_projection"]

_illumination = [
    "compute_illumination_surface",
    "correct_illumination_surface"]

_postprocess = [
    "identify_objects_in_region",
    "remove_transcription_site",
    "extract_cell",
    "extract_spots_from_frame",
    "summarize_extraction_results",
    "center_mask_coord",
    "from_boundaries_to_surface",
    "from_surface_to_boundaries",
    "from_binary_to_coord",
    "complete_coord_boundaries",
    "from_coord_to_frame",
    "from_coord_to_surface"]

_augmentation = [
    "augment_2d",
    "augment_2d_function",
    "augment_8_times",
    "augment_8_times_reversed"]

_quality = [
    "compute_focus"]


__all__ = (_utils + _io + _preprocess + _postprocess + _filter + _projection +
           _illumination + _augmentation + _quality)



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.stack.utils module.
"""

import os
import pytest
import tempfile

import bigfish.stack as stack

import numpy as np
import pandas as pd

from bigfish.stack.utils import fit_recipe
from bigfish.stack.utils import get_path_from_recipe
from bigfish.stack.utils import get_nb_element_per_dimension
from bigfish.stack.utils import count_nb_fov


# TODO add test for bigfish.stack.load_and_save_url
# TODO add test for bigfish.stack.check_hash
# TODO add test for bigfish.stack.compute_hash

# ### Test sanity check functions ###

def test_check_parameter():
    # define a function with different parameters to check
    def foo(a, b, c, d, e, f, g, h):
        stack.check_parameter(a=(list, type(None)),
                              b=str,
                              c=int,
                              d=float,
                              e=np.ndarray,
                              f=bool,
                              g=(pd.DataFrame, pd.Series),
                              h=pd.DataFrame)
        return True

    # test the consistency of the check function when it works...
    assert foo(a=[], b="bar", c=5, d=2.5, e=np.array([3, 6, 9]),
               f=True, g=pd.DataFrame(), h=pd.DataFrame())
    assert foo(a=None, b="", c=10, d=2.0, e=np.array([3, 6, 9]),
               f=False, g=pd.Series(), h=pd.DataFrame())

    # ... and when it should raise an error
    with pytest.raises(TypeError):
        foo(a=(), b="bar", c=5, d=2.5, e=np.array([3, 6, 9]),
            f=True, g=pd.DataFrame(), h=pd.DataFrame())
    with pytest.raises(TypeError):
        foo(a=[], b="bar", c=5.0, d=2.5, e=np.array([3, 6, 9]),
            f=True, g=pd.DataFrame(), h=pd.DataFrame())
    with pytest.raises(TypeError):
        foo(a=[], b="bar", c=5, d=2, e=np.array([3, 6, 9]),
            f=True, g=pd.DataFrame(), h=pd.DataFrame())
    with pytest.raises(TypeError):
        foo(a=[], b="bar", c=5, d=2.5, e=[3, 6, 9],
            f=True, g=pd.DataFrame(), h=pd.DataFrame())
    with pytest.raises(TypeError):
        foo(a=[], b="bar", c=5, d=2.5, e=np.zeros((3, 3)),
            f=True, g=pd.DataFrame(), h=pd.Series())


def test_check_df():
    # build a dataframe to test
    df = pd.DataFrame({"A": [3, 6, 9],
                       "B": [2.5, np.nan, 1.3],
                       "C": ["arthur", "florian", "thomas"],
                       "D": [True, True, False]})

    # test the consistency of the check function when it works...
    assert stack.check_df(df,
                          features=["A", "B", "C", "D"],
                          features_without_nan=["A", "C", "D"])
    assert stack.check_df(df,
                          features=["B", "A"],
                          features_without_nan=["C", "D", "A"])
    assert stack.check_df(df,
                          features=None,
                          features_without_nan=["A", "C", "D"])
    assert stack.check_df(df,
                          features=["A", "B", "C", "D"],
                          features_without_nan=None)

    # ... and when it should raise an error
    with pytest.raises(ValueError):
        stack.check_df(df,
                       features=["A", "B", "C", "D", "E"],
                       features_without_nan=["A", "C", "D"])
    with pytest.raises(ValueError):
        stack.check_df(df,
                       features=["A", "B", "C", "D"],
                       features_without_nan=["A", "B", "C", "D"])


def test_check_array():
    # build some arrays to test
    a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.float32)
    b = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.int64)
    c = np.array(([1, 2, 3]), dtype=np.float32)
    d = np.array(([1, 2, np.nan]), dtype=np.float32)

    # test number of dimensions
    assert stack.check_array(a, ndim=2)
    assert stack.check_array(b, ndim=[1, 2])
    assert stack.check_array(c, ndim=1)
    with pytest.raises(ValueError):
        stack.check_array(a, ndim=1)

    # test dtypes
    assert stack.check_array(a, dtype=np.float32)
    assert stack.check_array(b, dtype=np.int64)
    assert stack.check_array(c, dtype=[np.float32, np.int64])
    with pytest.raises(TypeError):
        stack.check_array(a, dtype=np.float64)

    # test missing values
    assert stack.check_array(a, allow_nan=False)
    assert stack.check_array(b, allow_nan=True)
    assert stack.check_array(d, allow_nan=True)
    with pytest.raises(ValueError):
        stack.check_array(d, allow_nan=False)


def test_check_range_value():
    # build some arrays to test
    a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.float32)

    # test the consistency of the check function when it works...
    assert stack.check_range_value(a, min_=1, max_=None)
    assert stack.check_range_value(a, min_=None, max_=9)

    # ... and when it should raise an error
    with pytest.raises(ValueError):
        stack.check_range_value(a, min_=2, max_=None)
    with pytest.raises(ValueError):
        stack.check_range_value(a, min_=None, max_=8)


# ### Test recipes ###

def test_check_recipe():
    # build a temporary directory with two files
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = os.path.join(tmp_dir, "experience_1_dapi_fov_1.tif")
        with open(path, 'w') as f:
            f.write("dapi file")
        path = os.path.join(tmp_dir, "experience_1_smfish_fov_1.tif")
        with open(path, 'w') as f:
            f.write("smFISH file")

        # test the consistency of the check function when it should work
        good_recipe_1 = {"fov": "fov_1",
                         "c": ["dapi", "smfish"],
                         "opt": "experience_1",
                         "ext": "tif",
                         "pattern": "opt_c_fov.ext"}
        assert stack.check_recipe(good_recipe_1, data_directory=None)
        assert stack.check_recipe(good_recipe_1, data_directory=tmp_dir)

        # case with a good recipe but when a file is missing
        good_recipe_2 = {"fov": "fov_1",
                         "c": ["dapi", "smfish", "cellmask"],
                         "opt": "experience_1",
                         "ext": "tif",
                         "pattern": "opt_c_fov.ext"}
        assert stack.check_recipe(good_recipe_2, data_directory=None)
        with pytest.raises(FileNotFoundError):
            stack.check_recipe(good_recipe_2, data_directory=tmp_dir)

        # cases without a 'pattern' key with a string value
        bad_recipe_1 = {"fov": "fov_1",
                        "c": ["dapi", "smfish"],
                        "opt": "experience_1",
                        "ext": "tif"}
        bad_recipe_2 = {"fov": "fov_1",
                        "c": ["dapi", "smfish"],
                        "opt": "experience_1",
                        "ext": "tif",
                        "pattern": ["opt_c_fov.ext"]}
        with pytest.raises(KeyError):
            stack.check_recipe(bad_recipe_1, data_directory=None)
        with pytest.raises(TypeError):
            stack.check_recipe(bad_recipe_2, data_directory=None)

        # case with a wrong pattern (repetitive key)
        bad_recipe_3 = {"fov": "fov_1",
                        "c": ["dapi", "smfish"],
                        "opt": "experience_1",
                        "ext": "tif",
                        "pattern": "opt_c_fov_fov.ext"}
        with pytest.raises(ValueError):
            stack.check_recipe(bad_recipe_3, data_directory=None)

        # case with wrong key or value
        bad_recipe_4 = {"fov": "fov_1",
                        "channel": ["dapi", "smfish"],
                        "optional": "experience_1",
                        "extension": "tif",
                        "pattern": "opt_c_fov.ext"}
        bad_recipe_5 = {"fov": "fov_1",
                        "c": ["dapi", "smfish"],
                        "opt": 1,
                        "ext": "tif",
                        "pattern": "opt_c_fov.ext"}
        with pytest.raises(KeyError):
            stack.check_recipe(bad_recipe_4, data_directory=None)
        with pytest.raises(TypeError):
            stack.check_recipe(bad_recipe_5, data_directory=None)


def test_fit_recipe():
    # build a recipe to fit
    good_recipe = {"fov": "fov_1",
                   "c": ["dapi", "smfish"],
                   "opt": "experience_1",
                   "ext": "tif",
                   "pattern": "opt_c_fov.ext"}

    # fit recipe
    new_recipe = fit_recipe(good_recipe)

    # all keys should be initialized in the new recipe, with a list or a string
    for key in ['fov', 'r', 'c', 'z']:
        assert key in new_recipe
        assert isinstance(new_recipe[key], list)
    for key in ['ext', 'opt']:
        assert key in new_recipe
        assert isinstance(new_recipe[key], str)
    assert 'pattern' in new_recipe
    assert isinstance(new_recipe['pattern'], str)

    # test that fitting an already fitted recipe does not change anything
    new_recip_bis = fit_recipe(new_recipe)
    assert new_recip_bis == new_recipe


def test_path_from_recipe():
    # build a temporary directory with one file
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = os.path.join(tmp_dir, "experience_1_dapi_fov_1.tif")
        with open(path, 'w') as f:
            f.write("dapi file")

        # build a recipe to read the file
        good_recipe = {"fov": "fov_1",
                       "c": "dapi",
                       "opt": "experience_1",
                       "ext": "tif",
                       "pattern": "opt_c_fov.ext"}

        # test the path
        path_dapi = get_path_from_recipe(good_recipe, tmp_dir, c=0)
        assert os.path.isfile(path_dapi)


def test_element_per_dimension():
    # build a recipe to test
    good_recipe = {"fov": "fov_1",
                   "c": ["dapi", "smfish"],
                   "opt": "experience_1",
                   "ext": "tif",
                   "pattern": "opt_c_fov.ext"}

    # test the number of elements to be stacked
    nb_r, nb_c, nb_z = get_nb_element_per_dimension(good_recipe)
    assert nb_r == 1
    assert nb_c == 2
    assert nb_z == 1


def test_nb_fov():
    # case when 'fov' key is a string
    good_recipe_1 = {"fov": "fov_1",
                     "c": ["dapi", "smfish"],
                     "opt": "experience_1",
                     "ext": "tif",
                     "pattern": "opt_c_fov.ext"}
    nb_fov = count_nb_fov(good_recipe_1)
    assert nb_fov == 1

    # case when 'fov' key is a list
    good_recipe_2 = {"fov": ["fov_1", "fov_2"],
                     "c": ["dapi", "smfish"],
                     "opt": "experience_1",
                     "ext": "tif",
                     "pattern": "opt_c_fov.ext"}
    nb_fov = count_nb_fov(good_recipe_2)
    assert nb_fov == 2

    # case when 'fov' key does not exist
    good_recipe_3 = {"c": ["dapi", "smfish"],
                     "opt": "experience_1",
                     "ext": "tif",
                     "pattern": "opt_c_fov.ext"}
    nb_fov = count_nb_fov(good_recipe_3)
    assert nb_fov == 1

    # case when the 'fov' key is not a string or a list
    with pytest.raises(TypeError):
        bad_recipe = {"fov": 1,
                      "c": ["dapi", "smfish"],
                      "opt": "experience_1",
                      "ext": "tif",
                      "pattern": "opt_c_fov.ext"}
        count_nb_fov(bad_recipe)


def test_check_datamap():
    # build a temporary directory with two files
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = os.path.join(tmp_dir, "experience_1_dapi_fov_1.tif")
        with open(path, 'w') as f:
            f.write("dapi file")
        path = os.path.join(tmp_dir, "experience_1_smfish_fov_1.tif")
        with open(path, 'w') as f:
            f.write("smFISH file")

        # test the consistency of the check function
        recipe = {"fov": "fov_1",
                  "c": ["dapi", "smfish"],
                  "opt": "experience_1",
                  "ext": "tif",
                  "pattern": "opt_c_fov.ext"}
        datamap = [(recipe, tmp_dir)]
        assert stack.check_datamap(datamap)
        datamap = [[recipe, tmp_dir]]
        assert stack.check_datamap(datamap)
        datamap = [(None, tmp_dir)]
        with pytest.raises(TypeError):
            stack.check_datamap(datamap)
        datamap = [(recipe, 3)]
        with pytest.raises(TypeError):
            stack.check_datamap(datamap)
        datamap = [(recipe, "/foo/bar")]
        with pytest.raises(NotADirectoryError):
            stack.check_datamap(datamap)
        datamap = [(recipe, tmp_dir, None)]
        with pytest.raises(ValueError):
            stack.check_datamap(datamap)


# ### Constants ###

def test_margin_value():
    # test margin value
    assert stack.get_margin_value() >= 2


def test_epsilon_float_32():
    # test epsilon value and dtype
    eps = stack.get_eps_float32()
    assert eps < 1e-5
    assert isinstance(eps, np.float32)



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.stack.augmentation module.
"""

import pytest

import numpy as np
import bigfish.stack as stack

from bigfish.stack.augmentation import _identity
from bigfish.stack.augmentation import _flip_h
from bigfish.stack.augmentation import _flip_v
from bigfish.stack.augmentation import _transpose
from bigfish.stack.augmentation import _transpose_inverse
from bigfish.stack.augmentation import _rotation_90
from bigfish.stack.augmentation import _rotation_180
from bigfish.stack.augmentation import _rotation_270

from numpy.testing import assert_array_equal


# toy image
x = np.array([[1, 0, 0, 0, 0],
              [0, 1, 0, 0, 0],
              [0, 1, 0, 0, 0],
              [0, 1, 1, 1, 0],
              [0, 0, 0, 0, 0]], dtype=np.uint8)


def test_identity():
    y = _identity(x)
    assert_array_equal(y, x)


def test_flip_h():
    # one channel
    y = _flip_h(x)
    expected_y = np.array([[0, 0, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [1, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)

    # multichannel
    xx = x[..., np.newaxis]
    yy = _flip_h(xx)
    expected_yy = expected_y[..., np.newaxis]
    assert_array_equal(yy, expected_yy)


def test_flip_v():
    # one channel
    y = _flip_v(x)
    expected_y = np.array([[0, 0, 0, 0, 1],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)

    # multichannel
    xx = x[..., np.newaxis]
    yy = _flip_v(xx)
    expected_yy = expected_y[..., np.newaxis]
    assert_array_equal(yy, expected_yy)


def test_transpose():
    # one channel
    y = _transpose(x)
    expected_y = np.array([[1, 0, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)

    # multichannel
    xx = x[..., np.newaxis]
    yy = _transpose(xx)
    expected_yy = expected_y[..., np.newaxis]
    assert_array_equal(yy, expected_yy)


def test_transpose_inverse():
    y = _transpose_inverse(x)
    expected_y = np.array([[0, 0, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 0, 1]], dtype=np.uint8)
    assert_array_equal(y, expected_y)


def test_rotation_90():
    y = _rotation_90(x)
    expected_y = np.array([[0, 0, 0, 0, 1],
                           [0, 1, 1, 1, 0],
                           [0, 1, 0, 0, 0],
                           [0, 1, 0, 0, 0],
                           [0, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)


def test_rotation_180():
    y = _rotation_180(x)
    expected_y = np.array([[0, 0, 0, 0, 0],
                           [0, 1, 1, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 0, 1]], dtype=np.uint8)
    assert_array_equal(y, expected_y)


def test_rotation_270():
    y = _rotation_270(x)
    expected_y = np.array([[0, 0, 0, 0, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 1, 0],
                           [0, 1, 1, 1, 0],
                           [1, 0, 0, 0, 0]], dtype=np.uint8)
    assert_array_equal(y, expected_y)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32,np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64,
    bool])
def test_augment_2d_dtype(dtype):
    x = np.array([[1, 0, 0, 0, 0],
                  [0, 1, 0, 0, 0],
                  [0, 1, 0, 0, 0],
                  [0, 1, 1, 1, 0],
                  [0, 0, 0, 0, 0]], dtype=dtype)
    y = stack.augment_2d(x)
    assert y.dtype == dtype



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.stack.preprocess module.
"""

import os
import pytest
import tempfile

import numpy as np
import bigfish.stack as stack

from numpy.testing import assert_array_equal


# ### Test stack building ###

def test_build_stacks_from_recipe():
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        # field of view 1
        test_nuc = np.zeros((8, 8, 8), dtype=np.uint8)
        test_cyt = np.zeros((8, 8, 8), dtype=np.uint8)
        test_rna = np.zeros((8, 8, 8), dtype=np.uint8)
        path_nuc = os.path.join(tmp_dir, "test_nuc_1.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt_1.tif")
        path_rna = os.path.join(tmp_dir, "test_rna_1.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # field of view 2
        test_nuc = np.zeros((5, 5, 5), dtype=np.uint16)
        test_cyt = np.zeros((5, 5, 5), dtype=np.uint16)
        test_rna = np.zeros((5, 5, 5), dtype=np.uint16)
        path_nuc = os.path.join(tmp_dir, "test_nuc_2.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt_2.tif")
        path_rna = os.path.join(tmp_dir, "test_rna_2.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # define recipe to read tensors
        recipe_1 ={"fov": ["1", "2"],
                   "c": ["nuc", "cyt", "rna"],
                   "opt": "test",
                   "ext": "tif",
                   "pattern": "opt_c_fov.ext"}

        # build tensor without prior information
        tensor = stack.build_stack(recipe_1, input_folder=tmp_dir)
        expected_tensor = np.zeros((1, 3, 8, 8, 8), dtype=np.uint8)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint8

        # build tensor with prior information
        tensor = stack.build_stack(recipe_1,
                                   input_folder=tmp_dir,
                                   input_dimension=3)
        expected_tensor = np.zeros((1, 3, 8, 8, 8), dtype=np.uint8)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint8

        # build tensors with different fields of view
        tensor = stack.build_stack(recipe_1,
                                   input_folder=tmp_dir,
                                   input_dimension=3,
                                   i_fov=0)
        expected_tensor = np.zeros((1, 3, 8, 8, 8), dtype=np.uint8)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint8
        tensor = stack.build_stack(recipe_1,
                                   input_folder=tmp_dir,
                                   input_dimension=3,
                                   i_fov=1)
        expected_tensor = np.zeros((1, 3, 5, 5, 5), dtype=np.uint16)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint16

        # wrong recipe
        recipe_wrong = {"fov": "test",
                        "c": ["nuc", "cyt", "rna"],
                        "ext": "tif",
                        "pattern": "fov_c.ext"}
        with pytest.raises(FileNotFoundError):
            stack.build_stack(recipe_wrong,
                              input_folder=tmp_dir,
                              input_dimension=3)

        # wrong path
        with pytest.raises(FileNotFoundError):
            stack.build_stack(recipe_1,
                              input_folder="/foo/bar",
                              input_dimension=3)


def test_build_stacks_from_datamap():
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        # field of view 1
        test_nuc = np.zeros((8, 8, 8), dtype=np.uint8)
        test_cyt = np.zeros((8, 8, 8), dtype=np.uint8)
        test_rna = np.zeros((8, 8, 8), dtype=np.uint8)
        path_nuc = os.path.join(tmp_dir, "test_nuc_1.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt_1.tif")
        path_rna = os.path.join(tmp_dir, "test_rna_1.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # field of view 2
        test_nuc = np.zeros((5, 5, 5), dtype=np.uint16)
        test_cyt = np.zeros((5, 5, 5), dtype=np.uint16)
        test_rna = np.zeros((5, 5, 5), dtype=np.uint16)
        path_nuc = os.path.join(tmp_dir, "test_nuc_2.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt_2.tif")
        path_rna = os.path.join(tmp_dir, "test_rna_2.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # define datamap to read  tensors
        recipe_1 = {"fov": ["1", "2"],
                    "c": ["nuc", "cyt", "rna"],
                    "opt": "test",
                    "ext": "tif",
                    "pattern": "opt_c_fov.ext"}
        recipe_2 = {"fov": "2",
                    "c": ["nuc", "cyt", "rna"],
                    "opt": "test",
                    "ext": "tif",
                    "pattern": "opt_c_fov.ext"}
        data_map = [(recipe_1, tmp_dir), (recipe_2, tmp_dir)]

        # build stacks from generator
        generator = stack.build_stacks(data_map, input_dimension=3)
        expected_tensors = [np.zeros((1, 3, 8, 8, 8), dtype=np.uint8),
                            np.zeros((1, 3, 5, 5, 5), dtype=np.uint16),
                            np.zeros((1, 3, 5, 5, 5), dtype=np.uint16)]
        for i, tensor in enumerate(generator):
            expected_tensor = expected_tensors[i]
            assert_array_equal(tensor, expected_tensor)
            assert tensor.dtype == expected_tensor.dtype

        # build stacks from generator with metadata
        generator = stack.build_stacks(data_map,
                                       input_dimension=3,
                                       return_origin=True)
        expected_tensors = [np.zeros((1, 3, 8, 8, 8), dtype=np.uint8),
                            np.zeros((1, 3, 5, 5, 5), dtype=np.uint16),
                            np.zeros((1, 3, 5, 5, 5), dtype=np.uint16)]
        expected_recipes = [recipe_1, recipe_1, recipe_2]
        expected_i_fov = [0, 1, 0]
        for i, (tensor, input_folder, recipe, i_fov) in enumerate(generator):
            expected_tensor = expected_tensors[i]
            assert_array_equal(tensor, expected_tensor)
            assert tensor.dtype == expected_tensor.dtype
            assert input_folder == tmp_dir
            assert recipe == expected_recipes[i]
            assert i_fov == expected_i_fov[i]

        # wrong datamap
        data_map = [(recipe_1, 3), (recipe_2, tmp_dir)]
        generator = stack.build_stacks(data_map, input_dimension=3)
        with pytest.raises(TypeError):
            next(generator)
        data_map = [(recipe_1, "foo/bar"), (recipe_2, tmp_dir)]
        generator = stack.build_stacks(data_map, input_dimension=3)
        with pytest.raises(NotADirectoryError):
            next(generator)


def test_build_stack_from_path():
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        # field of view
        test_nuc = np.zeros((8, 8, 8), dtype=np.uint8)
        test_cyt = np.zeros((8, 8, 8), dtype=np.uint8)
        test_rna = np.zeros((8, 8, 8), dtype=np.uint8)
        path_nuc = os.path.join(tmp_dir, "test_nuc.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt.tif")
        path_rna = os.path.join(tmp_dir, "test_rna.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # build tensor from paths
        paths = [path_nuc, path_cyt, path_rna]
        tensor = stack.build_stack_no_recipe(paths, input_dimension=3)
        expected_tensor = np.zeros((1, 3, 8, 8, 8), dtype=np.uint8)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint8

        # wrong paths
        paths = [path_nuc, path_cyt, "/foo/bar/test_rna.tif"]
        with pytest.raises(FileNotFoundError):
            stack.build_stack_no_recipe(paths, input_dimension=3)


# ### Test normalization ###

@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32,
    np.int8, np.int16, np.int32,
    np.float16, np.float32, np.float64])
def test_rescale(dtype):
    # build a 5x5 random matrix with a limited range of values
    tensor = np.random.randint(35, 105, 25).reshape((5, 5)).astype(dtype)

    # rescale tensor
    rescaled_tensor = stack.rescale(tensor)

    # test consistency of the function
    if dtype in [np.uint8, np.uint16, np.uint32, np.int8, np.int16, np.int32]:
        i = np.iinfo(dtype)
        min_, max_ = 0, i.max
    else:
        min_, max_ = 0, 1
    assert rescaled_tensor.min() == min_
    assert rescaled_tensor.max() == max_
    assert rescaled_tensor.dtype == dtype
    assert rescaled_tensor.shape == (5, 5)


def test_stretching():
    x = [[51, 51, 51], [102, 102, 102], [153, 153, 153]]

    # integer
    tensor = np.array(x).reshape((3, 3)).astype(np.uint16)
    tensor_rescaled = stack.rescale(tensor,
                                    channel_to_stretch=0,
                                    stretching_percentile=50)
    expected_tensor = np.array([[0, 0, 0],
                                [65535, 65535, 65535],
                                [65535, 65535, 65535]], dtype=np.uint16)
    assert_array_equal(tensor_rescaled, expected_tensor)

    # float
    tensor = np.array(x).reshape((3, 3)).astype(np.float32)
    rescaled_tensor = stack.rescale(tensor,
                                    channel_to_stretch=0,
                                    stretching_percentile=50)
    expected_tensor = np.array([[0., 0., 0.],
                                [1., 1., 1.],
                                [1., 1., 1.]], dtype=np.float32)
    assert_array_equal(rescaled_tensor, expected_tensor)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64])
def test_cast_uint8(dtype):
    # from integer to np.uint8
    if np.issubdtype(dtype, np.integer):
        x = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)
        tensor[2, 2] = np.iinfo(dtype).max

    # from float to np.uint8
    else:
        x = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 1.0]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)

    # cast in uint8
    if dtype in [np.uint8, np.int8]:
        tensor_uint8 = stack.cast_img_uint8(tensor)
    else:
        with pytest.warns(UserWarning):
            tensor_uint8 = stack.cast_img_uint8(tensor)
    assert tensor_uint8.dtype == np.uint8


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64])
def test_cast_uint16(dtype):
    # from integer to np.uint16
    if np.issubdtype(dtype, np.integer):
        x = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)
        tensor[2, 2] = np.iinfo(dtype).max

    # from float to np.uint16
    else:
        x = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 1.0]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)

    # cast in uint16
    if dtype in [np.uint8, np.int8, np.uint16, np.int16, np.float16]:
        tensor_uint16 = stack.cast_img_uint16(tensor)
    else:
        with pytest.warns(UserWarning):
            tensor_uint16 = stack.cast_img_uint16(tensor)
    assert tensor_uint16.dtype == np.uint16


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64])
def test_cast_float32(dtype):
    # from integer to np.float32
    if np.issubdtype(dtype, np.integer):
        x = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)
        tensor[2, 2] = np.iinfo(dtype).max

    # from float to np.float32
    else:
        x = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 1.0]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)

    # cast in float32
    if dtype == np.float64:
        with pytest.warns(UserWarning):
            tensor_float32 = stack.cast_img_float32(tensor)
    else:
        tensor_float32 = stack.cast_img_float32(tensor)
    assert tensor_float32.dtype == np.float32


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64])
def test_cast_float64(dtype):
    # from integer to np.float64
    if np.issubdtype(dtype, np.integer):
        x = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)
        tensor[2, 2] = np.iinfo(dtype).max

    # from float to np.float64
    else:
        x = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 1.0]]
        tensor = np.array(x).reshape((3, 3)).astype(dtype)

    # cast in float64
    tensor_float64 = stack.cast_img_float64(tensor)
    assert tensor_float64.dtype == np.float64



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions to plot 2-d pixel and coordinates images.
"""

import warnings

import bigfish.stack as stack

from .utils import save_plot, get_minmax_values, create_colormap

import matplotlib.pyplot as plt
import numpy as np

from skimage.segmentation import find_boundaries
from matplotlib.colors import ListedColormap
from matplotlib.patches import RegularPolygon


# ### General plot ###

def plot_yx(image, r=0, c=0, z=0, rescale=False, contrast=False,
            title=None, framesize=(8, 8), remove_frame=True, path_output=None,
            ext="png", show=True):
    """Plot the selected yx plan of the selected dimensions of an image.

    Parameters
    ----------
    image : np.ndarray
        A 2-d, 3-d, 4-d or 5-d image with shape (y, x), (z, y, x),
        (c, z, y, x) or (r, c, z, y, x) respectively.
    r : int
        Index of the round to keep.
    c : int
        Index of the channel to keep.
    z : int
        Index of the z slice to keep.
    rescale : bool
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool
        Contrast image.
    title : str
        Title of the image.
    framesize : tuple
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    remove_frame : bool
        Remove axes and frame.
    path_output : str
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.

    """
    # check parameters
    stack.check_array(image,
                      ndim=[2, 3, 4, 5],
                      dtype=[np.uint8, np.uint16, np.int64,
                             np.float32, np.float64,
                             bool])
    stack.check_parameter(r=int, c=int, z=int,
                          rescale=bool,
                          contrast=bool,
                          title=(str, type(None)),
                          framesize=tuple,
                          remove_frame=bool,
                          path_output=(str, type(None)),
                          ext=(str, list),
                          show=bool)

    # get the 2-d image
    if image.ndim == 2:
        xy_image = image
    elif image.ndim == 3:
        xy_image = image[z, :, :]
    elif image.ndim == 4:
        xy_image = image[c, z, :, :]
    else:
        xy_image = image[r, c, z, :, :]

    # plot
    if remove_frame:
        fig = plt.figure(figsize=framesize, frameon=False)
        ax = fig.add_axes([0, 0, 1, 1])
        ax.axis('off')
    else:
        plt.figure(figsize=framesize)
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        plt.imshow(xy_image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        plt.imshow(xy_image)
    else:
        if xy_image.dtype not in [np.int64, bool]:
            xy_image = stack.rescale(xy_image, channel_to_stretch=0)
        plt.imshow(xy_image)
    if title is not None and not remove_frame:
        plt.title(title, fontweight="bold", fontsize=25)
    if not remove_frame:
        plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


def plot_images(images, rescale=False, contrast=False, titles=None,
                framesize=(15, 10), remove_frame=True, path_output=None,
                ext="png", show=True):
    """Plot or subplot of 2-d images.

    Parameters
    ----------
    images : np.ndarray or List[np.ndarray]
        Images with shape (y, x).
    rescale : bool
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool
        Contrast image.
    titles : List[str]
        Titles of the subplots.
    framesize : tuple
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    remove_frame : bool
        Remove axes and frame.
    path_output : str
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.


    """
    # enlist image if necessary
    if isinstance(images, np.ndarray):
        images = [images]

    # check parameters
    stack.check_parameter(images=list,
                          rescale=bool,
                          contrast=bool,
                          titles=(str, list, type(None)),
                          framesize=tuple,
                          remove_frame=bool,
                          path_output=(str, type(None)),
                          ext=(str, list),
                          show=bool)
    for image in images:
        stack.check_array(image,
                          ndim=2,
                          dtype=[np.uint8, np.uint16, np.int64,
                                 np.float32, np.float64,
                                 bool])

    # we plot 3 images by row maximum
    nrow = int(np.ceil(len(images)/3))
    ncol = min(len(images), 3)

    # plot one image
    if len(images) == 1:
        if titles is not None:
            title = titles[0]
        else:
            title = None
        plot_yx(images[0],
                rescale=rescale,
                contrast=contrast,
                title=title,
                framesize=framesize,
                remove_frame=remove_frame,
                path_output=path_output,
                ext=ext,
                show=show)

        return

    # plot multiple images
    fig, ax = plt.subplots(nrow, ncol, figsize=framesize)

    # one row
    if len(images) in [2, 3]:
        for i, image in enumerate(images):
            if remove_frame:
                ax[i].axis("off")
            if not rescale and not contrast:
                vmin, vmax = get_minmax_values(image)
                ax[i].imshow(image, vmin=vmin, vmax=vmax)
            elif rescale and not contrast:
                ax[i].imshow(image)
            else:
                if image.dtype not in [np.int64, bool]:
                    image = stack.rescale(image, channel_to_stretch=0)
                ax[i].imshow(image)
            if titles is not None:
                ax[i].set_title(titles[i], fontweight="bold", fontsize=10)

    # several rows
    else:
        # we complete the row with empty frames
        r = nrow * 3 - len(images)
        images_completed = [image for image in images] + [None] * r

        for i, image in enumerate(images_completed):
            row = i // 3
            col = i % 3
            if image is None:
                ax[row, col].set_visible(False)
                continue
            if remove_frame:
                ax[row, col].axis("off")
            if not rescale and not contrast:
                vmin, vmax = get_minmax_values(image)
                ax[row, col].imshow(image, vmin=vmin, vmax=vmax)
            elif rescale and not contrast:
                ax[row, col].imshow(image)
            else:
                if image.dtype not in [np.int64, bool]:
                    image = stack.rescale(image, channel_to_stretch=0)
                ax[row, col].imshow(image)
            if titles is not None:
                ax[row, col].set_title(titles[i],
                                       fontweight="bold", fontsize=10)

    plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


# ### Segmentation plot ###

def plot_segmentation(image, mask, rescale=False, contrast=False, title=None,
                      framesize=(15, 10), remove_frame=True,
                      path_output=None, ext="png", show=True):
    """Plot result of a 2-d segmentation, with labelled instances if available.

    Parameters
    ----------
    image : np.ndarray
        A 2-d image with shape (y, x).
    mask : np.ndarray
        A 2-d image with shape (y, x).
    rescale : bool
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool
        Contrast image.
    title : str
        Title of the image.
    framesize : tuple
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    remove_frame : bool
        Remove axes and frame.
    path_output : str
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.

    """
    # check parameters
    stack.check_array(image,
                      ndim=2,
                      dtype=[np.uint8, np.uint16, np.int64,
                             np.float32, np.float64,
                             bool])
    stack.check_array(mask,
                      ndim=2,
                      dtype=[np.uint8, np.uint16, np.int64, bool])
    stack.check_parameter(rescale=bool,
                          contrast=bool,
                          title=(str, type(None)),
                          framesize=tuple,
                          remove_frame=bool,
                          path_output=(str, type(None)),
                          ext=(str, list))

    # plot
    fig, ax = plt.subplots(1, 3, sharex='col', figsize=framesize)

    # image
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        ax[0].imshow(image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        ax[0].imshow(image)
    else:
        if image.dtype not in [np.int64, bool]:
            image = stack.rescale(image, channel_to_stretch=0)
        ax[0].imshow(image)
    if title is not None:
        ax[0].set_title(title, fontweight="bold", fontsize=10)
    if remove_frame:
        ax[0].axis("off")

    # label
    ax[1].imshow(mask)
    if title is not None:
        ax[1].set_title("Segmentation", fontweight="bold", fontsize=10)
    if remove_frame:
        ax[1].axis("off")

    # superposition
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        ax[2].imshow(image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        ax[2].imshow(image)
    else:
        if image.dtype not in [np.int64, bool]:
            image = stack.rescale(image, channel_to_stretch=0)
        ax[2].imshow(image)
    masked = np.ma.masked_where(mask == 0, mask)
    ax[2].imshow(masked, cmap=ListedColormap(['red']), alpha=0.5)
    if title is not None:
        ax[2].set_title("Surface", fontweight="bold", fontsize=10)
    if remove_frame:
        ax[2].axis("off")

    plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


def plot_segmentation_boundary(image, cell_label=None, nuc_label=None,
                               rescale=False, contrast=False, title=None,
                               framesize=(10, 10), remove_frame=True,
                               path_output=None, ext="png", show=True):
    """Plot the boundary of the segmented objects.

    Parameters
    ----------
    image : np.ndarray
        A 2-d image with shape (y, x).
    cell_label : np.ndarray
        A 2-d image with shape (y, x).
    nuc_label : np.ndarray
        A 2-d image with shape (y, x).
    rescale : bool
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool
        Contrast image.
    title : str
        Title of the image.
    framesize : tuple
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    remove_frame : bool
        Remove axes and frame.
    path_output : str
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.

    """
    # check parameters
    stack.check_array(image,
                      ndim=2,
                      dtype=[np.uint8, np.uint16, np.int64,
                             np.float32, np.float64,
                             bool])
    if cell_label is not None:
        stack.check_array(cell_label,
                          ndim=2,
                          dtype=[np.uint8, np.uint16, np.int64, bool])
    if nuc_label is not None:
        stack.check_array(nuc_label,
                          ndim=2,
                          dtype=[np.uint8, np.uint16, np.int64, bool])
    stack.check_parameter(rescale=bool,
                          contrast=bool,
                          title=(str, type(None)),
                          framesize=tuple,
                          remove_frame=bool,
                          path_output=(str, type(None)),
                          ext=(str, list),
                          show=bool)

    # get boundaries
    cell_boundaries = None
    nuc_boundaries = None
    if cell_label is not None:
        cell_boundaries = find_boundaries(cell_label, mode='thick')
        cell_boundaries = np.ma.masked_where(cell_boundaries == 0,
                                             cell_boundaries)
    if nuc_label is not None:
        nuc_boundaries = find_boundaries(nuc_label, mode='thick')
        nuc_boundaries = np.ma.masked_where(nuc_boundaries == 0,
                                            nuc_boundaries)

    # plot
    if remove_frame:
        fig = plt.figure(figsize=framesize, frameon=False)
        ax = fig.add_axes([0, 0, 1, 1])
        ax.axis('off')
    else:
        plt.figure(figsize=framesize)
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        plt.imshow(image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        plt.imshow(image)
    else:
        if image.dtype not in [np.int64, bool]:
            image = stack.rescale(image, channel_to_stretch=0)
        plt.imshow(image)
    if cell_label is not None:
        plt.imshow(cell_boundaries, cmap=ListedColormap(['red']))
    if nuc_label is not None:
        plt.imshow(nuc_boundaries, cmap=ListedColormap(['blue']))
    if title is not None and not remove_frame:
        plt.title(title, fontweight="bold", fontsize=25)
    if not remove_frame:
        plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


def plot_segmentation_diff(image, mask_pred, mask_gt, rescale=False,
                           contrast=False, title=None, framesize=(15, 10),
                           remove_frame=True, path_output=None, ext="png",
                           show=True):
    """Plot segmentation results along with ground truth to compare.

    Parameters
    ----------
    image : np.ndarray, np.uint, np.int, np.float or bool
        Image with shape (y, x).
    mask_pred : np.ndarray, np.uint, np.int or np.float
        Image with shape (y, x).
    mask_gt : np.ndarray, np.uint, np.int or np.float
        Image with shape (y, x).
    rescale : bool
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool
        Contrast image.
    title : str or None
        Title of the plot.
    framesize : tuple
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    remove_frame : bool
        Remove axes and frame.
    path_output : str or None
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.

    """
    # check parameters
    stack.check_parameter(rescale=bool,
                          contrast=bool,
                          title=(str, type(None)),
                          framesize=tuple,
                          remove_frame=bool,
                          path_output=(str, type(None)),
                          ext=(str, list),
                          show=bool)
    stack.check_array(image,
                      ndim=2,
                      dtype=[np.uint8, np.uint16, np.int64,
                             np.float32, np.float64,
                             bool])
    stack.check_array(mask_pred,
                      ndim=2,
                      dtype=[np.uint8, np.uint16, np.int32, np.int64,
                             np.float32, np.float64,
                             bool])
    stack.check_array(mask_gt,
                      ndim=2,
                      dtype=[np.uint8, np.uint16, np.int32, np.int64,
                             np.float32, np.float64,
                             bool])

    # plot multiple images
    fig, ax = plt.subplots(1, 3, figsize=framesize)

    # image
    if remove_frame:
        ax[0].axis("off")
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        ax[0].imshow(image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        ax[0].imshow(image)
    else:
        if image.dtype not in [np.int64, bool]:
            image = stack.rescale(image, channel_to_stretch=0)
        ax[0].imshow(image)
    if title is None:
        ax[0].set_title("", fontweight="bold", fontsize=10)
    else:
        ax[0].set_title(title, fontweight="bold", fontsize=10)

    # build colormap
    cmap = create_colormap()

    # prediction
    im_mask_pred = np.ma.masked_where(mask_pred == 0, mask_pred)
    if remove_frame:
        ax[1].axis("off")
    ax[1].imshow(im_mask_pred, cmap=cmap)
    ax[1].set_title("Prediction", fontweight="bold", fontsize=10)

    # ground truth
    im_mask_gt = np.ma.masked_where(mask_gt == 0, mask_gt)
    if remove_frame:
        ax[2].axis("off")
    ax[2].imshow(im_mask_gt, cmap=cmap)
    ax[2].set_title("Ground truth", fontweight="bold", fontsize=10)

    plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


# ### Detection plot ###

# TODO allow textual annotations
def plot_detection(image, spots, shape="circle", radius=3, color="red",
                   linewidth=1, fill=False, rescale=False, contrast=False,
                   title=None, framesize=(15, 10), remove_frame=True,
                   path_output=None, ext="png", show=True):
    """Plot detected spots and foci on a 2-d image.

    Parameters
    ----------
    image : np.ndarray
        A 2-d image with shape (y, x).
    spots : List[np.ndarray] or np.ndarray
        Array with coordinates and shape (nb_spots, 3) or (nb_spots, 2). To
        plot different kind of detected spots with different symbols, use a
        list of arrays.
    shape : List[str] or str
        List of symbols used to localized the detected spots in the image,
        among `circle`, `square` or `polygon`. One symbol per array in `spots`.
        If `shape` is a string, the same symbol is used for every elements of
        'spots'.
    radius : List[int or float] or int or float
        List of yx radii of the detected spots. One radius per array in
        `spots`. If `radius` is a scalar, the same value is applied for every
        elements of `spots`.
    color : List[str] or str
        List of colors of the detected spots. One color per array in `spots`.
        If `color` is a string, the same color is applied for every elements
        of `spots`.
    linewidth : List[int] or int
        List of widths or width of the border symbol. One integer per array
        in `spots`. If `linewidth` is an integer, the same width is applied
        for every elements of `spots`.
    fill : List[bool] or bool
        List of boolean to fill the symbol the detected spots.
    rescale : bool
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool
        Contrast image.
    title : str
        Title of the image.
    framesize : tuple
        Size of the frame used to plot with ``plt.figure(figsize=framesize``.
    remove_frame : bool
        Remove axes and frame.
    path_output : str
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.

    """
    # check parameters
    stack.check_array(image,
                      ndim=2,
                      dtype=[np.uint8, np.uint16, np.int64,
                             np.float32, np.float64])
    stack.check_parameter(spots=(list, np.ndarray),
                          shape=(list, str),
                          radius=(list, int, float),
                          color=(list, str),
                          linewidth=(list, int),
                          fill=(list, bool),
                          rescale=bool,
                          contrast=bool,
                          title=(str, type(None)),
                          framesize=tuple,
                          remove_frame=bool,
                          path_output=(str, type(None)),
                          ext=(str, list),
                          show=bool)
    if isinstance(spots, list):
        for spots_ in spots:
            stack.check_array(spots_, ndim=2, dtype=[np.int64, np.float64])
    else:
        stack.check_array(spots, ndim=2, dtype=[np.int64, np.float64])

    # enlist and format parameters
    if not isinstance(spots, list):
        spots = [spots]
    n = len(spots)
    if not isinstance(shape, list):
        shape = [shape] * n
    elif isinstance(shape, list) and len(shape) != n:
        raise ValueError("If 'shape' is a list, it should have the same "
                         "number of items than spots ({0}).".format(n))
    if not isinstance(radius, list):
        radius = [radius] * n
    elif isinstance(radius, list) and len(radius) != n:
        raise ValueError("If 'radius' is a list, it should have the same "
                         "number of items than spots ({0}).".format(n))
    if not isinstance(color, list):
        color = [color] * n
    elif isinstance(color, list) and len(color) != n:
        raise ValueError("If 'color' is a list, it should have the same "
                         "number of items than spots ({0}).".format(n))
    if not isinstance(linewidth, list):
        linewidth = [linewidth] * n
    elif isinstance(linewidth, list) and len(linewidth) != n:
        raise ValueError("If 'linewidth' is a list, it should have the same "
                         "number of items than spots ({0}).".format(n))
    if not isinstance(fill, list):
        fill = [fill] * n
    elif isinstance(fill, list) and len(fill) != n:
        raise ValueError("If 'fill' is a list, it should have the same "
                         "number of items than spots ({0}).".format(n))

    # plot
    fig, ax = plt.subplots(1, 2, sharex='col', figsize=framesize)

    # image
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        ax[0].imshow(image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        ax[0].imshow(image)
    else:
        if image.dtype not in [np.int64, bool]:
            image = stack.rescale(image, channel_to_stretch=0)
        ax[0].imshow(image)

    # spots
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        ax[1].imshow(image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        ax[1].imshow(image)
    else:
        if image.dtype not in [np.int64, bool]:
            image = stack.rescale(image, channel_to_stretch=0)
        ax[1].imshow(image)

    for i, coordinates in enumerate(spots):

        # get 2-d coordinates
        if coordinates.shape[1] == 3:
            coordinates_2d = coordinates[:, 1:]
        else:
            coordinates_2d = coordinates

        # plot symbols
        for y, x in coordinates_2d:
            x = _define_patch(x, y, shape[i], radius[i], color[i],
                              linewidth[i], fill[i])
            ax[1].add_patch(x)

    # titles and frames
    if title is not None:
        ax[0].set_title(title, fontweight="bold", fontsize=10)
        ax[1].set_title("Detection results", fontweight="bold", fontsize=10)
    if remove_frame:
        ax[0].axis("off")
        ax[1].axis("off")
    plt.tight_layout()

    # output
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


def _define_patch(x, y, shape, radius, color, linewidth, fill):
    """Define a matplotlib.patches to plot.

    Parameters
    ----------
    x : int or float
        Coordinate x for the patch center.
    y : int or float
        Coordinate y for the patch center.
    shape : str
        Shape of the patch to define (among `circle`, `square` or `polygon`)
    radius : int or float
        Radius of the patch.
    color : str
        Color of the patch.
    linewidth : int
        Width of the patch border.
    fill : bool
        Make the patch shape empty or not.

    Returns
    -------
    x : matplotlib.patches object
        Geometric form to add to a plot.

    """
    # circle
    if shape == "circle":
        x = plt.Circle((x, y), radius,
                       color=color,
                       linewidth=linewidth,
                       fill=fill)
    # square
    elif shape == "square":
        x = plt.Rectangle((x, y), radius, radius,
                          color=color,
                          linewidth=linewidth,
                          fill=fill)
    # polygon
    elif shape == "polygon":
        x = RegularPolygon((x, y), 5, radius,
                           color=color,
                           linewidth=linewidth,
                           fill=fill)
    else:
        warnings.warn("shape should take a value among 'circle', 'square' or "
                      "'polygon', but not {0}".format(shape), UserWarning)

    return x


def plot_reference_spot(reference_spot, rescale=False, contrast=False,
                        title=None, framesize=(8, 8), remove_frame=True,
                        path_output=None, ext="png", show=True):
    """Plot the selected yx plan of the selected dimensions of an image.

    Parameters
    ----------
    reference_spot : np.ndarray
        Spot image with shape (z, y, x) or (y, x).
    rescale : bool
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool
        Contrast image.
    title : str
        Title of the image.
    framesize : tuple
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    remove_frame : bool
        Remove axes and frame.
    path_output : str
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.

    """
    # check parameters
    stack.check_array(reference_spot,
                      ndim=[2,  3],
                      dtype=[np.uint8, np.uint16, np.int64,
                             np.float32, np.float64])
    stack.check_parameter(rescale=bool,
                          contrast=bool,
                          title=(str, type(None)),
                          framesize=tuple,
                          remove_frame=bool,
                          path_output=(str, type(None)),
                          ext=(str, list),
                          show=bool)

    # project spot in 2-d if necessary
    if reference_spot.ndim == 3:
        reference_spot = stack.maximum_projection(reference_spot)

    # plot reference spot
    if remove_frame:
        fig = plt.figure(figsize=framesize, frameon=False)
        ax = fig.add_axes([0, 0, 1, 1])
        ax.axis('off')
    else:
        plt.figure(figsize=framesize)
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(reference_spot)
        plt.imshow(reference_spot, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        plt.imshow(reference_spot)
    else:
        if reference_spot.dtype not in [np.int64, bool]:
            reference_spot = stack.rescale(reference_spot, channel_to_stretch=0)
        plt.imshow(reference_spot)
    if title is not None and not remove_frame:
        plt.title(title, fontweight="bold", fontsize=25)
    if not remove_frame:
        plt.tight_layout()
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


# ### Individual cell plot ###

def plot_cell(ndim, cell_coord=None, nuc_coord=None, rna_coord=None,
              foci_coord=None, other_coord=None, image=None, cell_mask=None,
              nuc_mask=None, title=None, remove_frame=True, rescale=False,
              contrast=False, framesize=(15, 10), path_output=None, ext="png",
              show=True):
    """
    Plot image and coordinates extracted for a specific cell.

    Parameters
    ----------
    ndim : int
        Number of spatial dimensions to consider in the coordinates (2 or 3).
    cell_coord : np.ndarray, np.int64
        Coordinates of the cell border with shape (nb_points, 2). If None,
        coordinate representation of the cell is not shown.
    nuc_coord : np.ndarray, np.int64
        Coordinates of the nucleus border with shape (nb_points, 2).
    rna_coord : np.ndarray, np.int64
        Coordinates of the detected spots with shape (nb_spots, 4) or
        (nb_spots, 3). One coordinate per dimension (zyx or yx dimensions)
        plus the index of the cluster assigned to the spot. If no cluster was
        assigned, value is -1. If only coordinates of spatial dimensions are
        available, only centroid of foci can be shown.
    foci_coord : np.ndarray, np.int64
        Array with shape (nb_foci, 5) or (nb_foci, 4). One coordinate per
        dimension for the foci centroid (zyx or yx dimensions), the number of
        spots detected in the foci and its index.
    other_coord : np.ndarray, np.int64
        Coordinates of the detected elements with shape (nb_elements, 3) or
        (nb_elements, 2). One coordinate per dimension (zyx or yx dimensions).
    image : np.ndarray, np.uint
        Original image of the cell with shape (y, x). If None, original image
        of the cell is not shown.
    cell_mask : np.ndarray, np.uint
        Mask of the cell.
    nuc_mask : np.ndarray, np.uint
        Mask of the nucleus.
    title : str or None
        Title of the image.
    remove_frame : bool
        Remove axes and frame.
    rescale : bool
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool
        Contrast image.
    framesize : tuple
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    path_output : str or None
        Path to save the image (without extension).
    ext : str or List[str]
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool
        Show the figure or not.

    """
    if cell_coord is None and image is None:
        return

    # check parameters
    if cell_coord is not None:
        stack.check_array(cell_coord, ndim=2, dtype=np.int64)
    if nuc_coord is not None:
        stack.check_array(nuc_coord, ndim=2, dtype=np.int64)
    if rna_coord is not None:
        stack.check_array(rna_coord, ndim=2, dtype=np.int64)
    if foci_coord is not None:
        stack.check_array(foci_coord, ndim=2, dtype=np.int64)
    if other_coord is not None:
        stack.check_array(other_coord, ndim=2, dtype=np.int64)
    if image is not None:
        stack.check_array(image, ndim=2,
                          dtype=[np.uint8, np.uint16, np.int64,
                                 np.float32, np.float64])
    if cell_mask is not None:
        stack.check_array(cell_mask,
                          ndim=2,
                          dtype=[np.uint8, np.uint16, np.int64, bool])
    if nuc_mask is not None:
        stack.check_array(nuc_mask,
                          ndim=2,
                          dtype=[np.uint8, np.uint16, np.int64, bool])
    stack.check_parameter(ndim=int,
                          title=(str, type(None)),
                          remove_frame=bool,
                          rescale=bool,
                          contrast=bool,
                          framesize=tuple,
                          path_output=(str, type(None)),
                          ext=(str, list))

    # plot original image and coordinate representation
    if cell_coord is not None and image is not None:
        fig, ax = plt.subplots(1, 2, figsize=framesize)

        # original image
        if not rescale and not contrast:
            vmin, vmax = get_minmax_values(image)
            ax[0].imshow(image, vmin=vmin, vmax=vmax)
        elif rescale and not contrast:
            ax[0].imshow(image)
        else:
            if image.dtype not in [np.int64, bool]:
                image = stack.rescale(image, channel_to_stretch=0)
            ax[0].imshow(image)
        if cell_mask is not None:
            cell_boundaries = stack.from_surface_to_boundaries(cell_mask)
            cell_boundaries = np.ma.masked_where(cell_boundaries == 0,
                                                 cell_boundaries)
            ax[0].imshow(cell_boundaries, cmap=ListedColormap(['red']))
        if nuc_mask is not None:
            nuc_boundaries = stack.from_surface_to_boundaries(nuc_mask)
            nuc_boundaries = np.ma.masked_where(nuc_boundaries == 0,
                                                nuc_boundaries)
            ax[0].imshow(nuc_boundaries, cmap=ListedColormap(['blue']))

        # coordinate image
        ax[1].plot(cell_coord[:, 1], cell_coord[:, 0],
                   c="black", linewidth=2)
        if nuc_coord is not None:
            ax[1].plot(nuc_coord[:, 1], nuc_coord[:, 0],
                       c="steelblue", linewidth=2)
        if rna_coord is not None:
            ax[1].scatter(rna_coord[:, ndim - 1], rna_coord[:, ndim - 2],
                          s=25, c="firebrick", marker=".")
        if foci_coord is not None:
            for foci in foci_coord:
                ax[1].text(foci[ndim-1] + 5, foci[ndim-2] - 5, str(foci[ndim]),
                           color="darkorange", size=20)
            # case where we know which rna belong to a foci
            if rna_coord.shape[1] == ndim + 1:
                foci_indices = foci_coord[:, ndim + 1]
                mask_rna_in_foci = np.isin(rna_coord[:, ndim], foci_indices)
                rna_in_foci_coord = rna_coord[mask_rna_in_foci, :].copy()
                ax[1].scatter(rna_in_foci_coord[:, ndim - 1],
                              rna_in_foci_coord[:, ndim - 2],
                              s=25, c="darkorange", marker=".")
            # case where we only know the foci centroid
            else:
                ax[1].scatter(foci_coord[:, ndim - 1], foci_coord[:, ndim - 2],
                              s=40, c="darkorange", marker="o")
        if other_coord is not None:
            ax[1].scatter(other_coord[:, ndim - 1], other_coord[:, ndim - 2],
                          s=25, c="forestgreen", marker="D")

        # titles and frames
        _, _, min_y, max_y = ax[1].axis()
        ax[1].set_ylim(max_y, min_y)
        ax[1].use_sticky_edges = True
        ax[1].margins(0.01, 0.01)
        ax[1].axis('scaled')
        if remove_frame:
            ax[0].axis("off")
            ax[1].axis("off")
        if title is not None:
            ax[0].set_title("Original image ({0})".format(title),
                            fontweight="bold", fontsize=10)
            ax[1].set_title("Coordinate representation ({0})".format(title),
                            fontweight="bold", fontsize=10)
        plt.tight_layout()

        # output
        if path_output is not None:
            save_plot(path_output, ext)
        if show:
            plt.show()
        else:
            plt.close()

    # plot coordinate representation only
    elif cell_coord is not None and image is None:
        if remove_frame:
            fig = plt.figure(figsize=framesize, frameon=False)
            ax = fig.add_axes([0, 0, 1, 1])
            ax.axis('off')
        else:
            plt.figure(figsize=framesize)

        # coordinate image
        plt.plot(cell_coord[:, 1], cell_coord[:, 0], c="black", linewidth=2)
        if nuc_coord is not None:
            plt.plot(nuc_coord[:, 1], nuc_coord[:, 0],
                     c="steelblue", linewidth=2)
        if rna_coord is not None:
            plt.scatter(rna_coord[:, ndim - 1], rna_coord[:, ndim - 2],
                        s=25, c="firebrick", marker=".")
        if foci_coord is not None:
            for foci in foci_coord:
                plt.text(foci[ndim-1] + 5, foci[ndim-2] - 5, str(foci[ndim]),
                         color="darkorange", size=20)
            # case where we know which rna belong to a foci
            if rna_coord.shape[1] == ndim + 1:
                foci_indices = foci_coord[:, ndim + 1]
                mask_rna_in_foci = np.isin(rna_coord[:, ndim], foci_indices)
                rna_in_foci_coord = rna_coord[mask_rna_in_foci, :].copy()
                plt.scatter(rna_in_foci_coord[:, ndim - 1],
                            rna_in_foci_coord[:, ndim - 2],
                            s=25, c="darkorange", marker=".")
            # case where we only know the foci centroid
            else:
                plt.scatter(foci_coord[:, ndim - 1], foci_coord[:, ndim - 2],
                            s=40, c="darkorange", marker="o")
        if other_coord is not None:
            plt.scatter(other_coord[:, ndim - 1], other_coord[:, ndim - 2],
                        s=25, c="forestgreen", marker="D")

        # titles and frames
        _, _, min_y, max_y = plt.axis()
        plt.ylim(max_y, min_y)
        plt.use_sticky_edges = True
        plt.margins(0.01, 0.01)
        plt.axis('scaled')
        if title is not None:
            plt.title("Coordinate representation ({0})".format(title),
                      fontweight="bold", fontsize=10)
        if not remove_frame:
            plt.tight_layout()

        # output
        if path_output is not None:
            save_plot(path_output, ext)
        if show:
            plt.show()
        else:
            plt.close()

    # plot original image only
    elif cell_coord is None and image is not None:
        plot_segmentation_boundary(
            image=image, cell_label=cell_mask, nuc_label=nuc_mask,
            rescale=rescale, contrast=contrast, title=title,
            framesize=framesize, remove_frame=remove_frame,
            path_output=path_output, ext=ext, show=show)



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
The bigfish package.
"""


# keep a MAJOR.MINOR.PATCH format
# MAJOR: major API changes
# MINOR: new features
# PATCH: backwards compatible bug fixes
# MAJOR.MINOR.PATCHdev means a version under development
__version__ = "0.5.0"



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Utility functions for bigfish.stack subpackage.
"""

import os
import re
import copy
import inspect
import hashlib

import numpy as np
import pandas as pd

from urllib.request import urlretrieve


# ### Sanity checks dataframe ###

def check_df(df, features=None, features_without_nan=None):
    """Full safety check of a dataframe.

    Parameters
    ----------
    df : pd.DataFrame or pd.Series
        Dataframe or Series to check.
    features : List[str]
        Names of the expected features.
    features_without_nan : List[str]
        Names of the features to check for the missing values

    Returns
    -------
    _ : bool
        Assert if the dataframe is well formatted.

    """
    # check parameters
    check_parameter(df=(pd.DataFrame, pd.Series),
                    features=(list, type(None)),
                    features_without_nan=(list, type(None)))

    # check features
    if features is not None:
        _check_features_df(df, features)

    # check NaN values
    if features_without_nan is not None:
        _check_features_df(df, features_without_nan)
        _check_nan_df(df, features_without_nan)

    return True


def _check_features_df(df, features):
    """Check that the dataframe contains expected features.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to check.
    features : List[str]
        Names of the expected features.

    """
    # check columns
    if not set(features).issubset(df.columns):
        raise ValueError("The dataframe does not seem to have the right "
                         "features. {0} instead of {1}"
                         .format(list(df.columns.values), features))


def _check_nan_df(df, features_to_check=None):
    """Check specific columns of the dataframe do not have any missing values.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to check.
    features_to_check : List[str]
        Names of the checked features.

    """
    # count NaN
    nan_count = df.isnull().sum()

    # for the full dataframe...
    if features_to_check is None:
        x = nan_count.sum()
        if x > 0:
            raise ValueError("The dataframe has {0} NaN values.".format(x))

    # ...or for some features
    else:
        nan_count = nan_count[features_to_check]
        x = nan_count.sum()
        if x > 0:
            raise ValueError("The dataframe has {0} NaN values for the "
                             "requested features: \n{1}.".format(x, nan_count))


# ### Sanity checks array ###

def check_array(array, ndim=None, dtype=None, allow_nan=True):
    """Full safety check of an array.

    Parameters
    ----------
    array : np.ndarray
        Array to check.
    ndim : int or List[int]
        Number of dimensions expected.
    dtype : type or List[type]
        Types expected.
    allow_nan : bool
        Allow NaN values or not.

    Returns
    -------
    _ : bool
        Assert if the array is well formatted.

    """
    # check parameters
    check_parameter(array=np.ndarray,
                    ndim=(int, list, type(None)),
                    dtype=(type, list, type(None)),
                    allow_nan=bool)

    # check the dtype
    if dtype is not None:
        _check_dtype_array(array, dtype)

    # check the number of dimension
    if ndim is not None:
        _check_dim_array(array, ndim)

    # check NaN
    if not allow_nan:
        _check_nan_array(array)

    return True


def _check_dtype_array(array, dtype):
    """Check that a np.ndarray has the right dtype.

    Parameters
    ----------
    array : np.ndarray
        Array to check
    dtype : type or List[type]
        Type expected.

    """
    # enlist the dtype expected
    if isinstance(dtype, type):
        dtype = [dtype]

    # TODO simplify
    # check the dtype of the array
    error = True
    for dtype_expected in dtype:
        if array.dtype == dtype_expected:
            error = False
            break

    if error:
        raise TypeError("{0} is not supported yet. Use one of those dtypes "
                        "instead: {1}.".format(array.dtype, dtype))


def _check_dim_array(array, ndim):
    """Check that the array has the right number of dimensions.

    Parameters
    ----------
    array : np.ndarray
        Array to check.
    ndim : int or List[int]
        Number of dimensions expected

    """
    # enlist the number of expected dimensions
    if isinstance(ndim, int):
        ndim = [ndim]

    # check the number of dimensions of the array
    if array.ndim not in ndim:
        raise ValueError("Array can't have {0} dimension(s). Expected "
                         "dimensions are: {1}.".format(array.ndim, ndim))


def _check_nan_array(array):
    """Check that the array does not have missing values.

    Parameters
    ----------
    array : np.ndarray
        Array to check.

    """
    # count nan
    mask = np.isnan(array)
    x = mask.sum()

    # check the NaN values of the array
    if x > 0:
        raise ValueError("Array has {0} NaN values.".format(x))


def check_range_value(array, min_=None, max_=None):
    """Check the support of the array.

    Parameters
    ----------
    array : np.ndarray
        Array to check.
    min_ : int
        Minimum value allowed.
    max_ : int
        Maximum value allowed.

    Returns
    -------
    _ : bool
        Assert if the array has the right range of values.

    """
    # check lowest and highest bounds
    if min_ is not None and array.min() < min_:
        raise ValueError("The array should have a lower bound of {0}, but its "
                         "minimum value is {1}.".format(min_, array.min()))
    if max_ is not None and array.max() > max_:
        raise ValueError("The array should have an upper bound of {0}, but "
                         "its maximum value is {1}.".format(max_, array.max()))

    return True


# ### Recipe management (sanity checks, fitting) ###

def check_recipe(recipe, data_directory=None):
    """Check and validate a recipe.

    Checking a recipe consists in validating its filename pattern and the
    content of the dictionary.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Can only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.
    data_directory : str
        Path of the directory with the files describes in the recipe. If it is
        provided, the function check that the files exist.

    Returns
    -------
    _ : bool
        Assert if the recipe is well formatted.

    """
    # check parameters
    check_parameter(recipe=dict,
                    data_directory=(str, type(None)))

    # check the filename pattern
    if "pattern" not in recipe:
        raise KeyError("A recipe should have a filename pattern "
                       "('pattern' keyword).")
    recipe_pattern = recipe["pattern"]
    if not isinstance(recipe_pattern, str):
        raise TypeError("'pattern' should be a string, not a {0}."
                        .format(type(recipe_pattern)))

    # count the different dimensions to combinate in the recipe (among
    # 'fov', 'r', 'c' and 'z')
    dimensions = re.findall("fov|r|c|z", recipe_pattern)

    # each dimension can only appear once in the filename pattern
    if len(dimensions) != len(set(dimensions)):
        raise ValueError("The pattern used in recipe is wrong, a dimension "
                         "appears several times: {0}".format(recipe_pattern))

    # check keys and values of the recipe
    for key, value in recipe.items():
        if key not in ['fov', 'r', 'c', 'z', 'ext', 'opt', 'pattern']:
            raise KeyError("The recipe can only contain the keys 'fov', 'r', "
                           "'c', 'z', 'ext', 'opt' or 'pattern'. Not '{0}'."
                           .format(key))
        if not isinstance(value, (list, str)):
            raise TypeError("A recipe can only contain lists or strings, "
                            "not {0}.".format(type(value)))

    # check that requested files exist
    if data_directory is not None:
        if not os.path.isdir(data_directory):
            raise NotADirectoryError("Directory does not exist: {0}"
                                     .format(data_directory))
        recipe = fit_recipe(recipe)
        nb_r, nb_c, nb_z = get_nb_element_per_dimension(recipe)
        nb_fov = count_nb_fov(recipe)
        for fov in range(nb_fov):
            for r in range(nb_r):
                for c in range(nb_c):
                    for z in range(nb_z):
                        path = get_path_from_recipe(recipe, data_directory,
                                                    fov=fov, r=r, c=c, z=z)
                        if not os.path.isfile(path):
                            raise FileNotFoundError("File does not exist: {0}"
                                                    .format(path))

    return True


def fit_recipe(recipe):
    """Fit a recipe.

    Fitting a recipe consists in wrapping every values of `fov`, `r`, `c` and
    `z` in a list (an empty one if necessary). Values for `ext` and `opt` are
    also initialized.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Can only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.

    Returns
    -------
    new_recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` and `opt`, initialized if
        necessary.

    """
    # check parameters
    check_parameter(recipe=dict)

    # initialize recipe
    new_recipe = copy.deepcopy(recipe)

    # initialize and fit the dimensions 'fov', 'r', 'c' and 'z'
    for key in ['fov', 'r', 'c', 'z']:
        if key not in new_recipe:
            new_recipe[key] = [None]
        value = new_recipe[key]
        if isinstance(value, str):
            new_recipe[key] = [value]

    # initialize the dimensions 'ext', 'opt'
    for key in ['ext', 'opt']:
        if key not in new_recipe:
            new_recipe[key] = ""

    return new_recipe


def _is_recipe_fitted(recipe):
    """Check if a recipe is ready to be used.

    Fitting a recipe consists in wrapping every values of `fov`, `r`, `c` and
    `z` in a list (an empty one if necessary). Values for `ext` and `opt` are
    also initialized.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Can only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.

    Returns
    -------
    _ : bool
        Indicates if the recipe is fitted or not

    """
    # all keys should be initialized in the new recipe, with a list or a string
    for key in ['fov', 'r', 'c', 'z']:
        if key not in recipe or not isinstance(recipe[key], list):
            return False
    for key in ['ext', 'opt']:
        if key not in recipe or not isinstance(recipe[key], str):
            return False
    if 'pattern' not in recipe or not isinstance(recipe['pattern'], str):
        return False

    return True


def get_path_from_recipe(recipe, input_folder, fov=0, r=0, c=0, z=0):
    """Build the path of a file from a recipe and the indices of specific
    elements.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.
    input_folder : str
        Path of the folder containing the images.
    fov : int
        Index of the `fov` element in the recipe to use in the filename.
    r : int
        Index of the `r` element in the recipe to use in the filename.
    c : int
        Index of the `c` element in the recipe to use in the filename.
    z : int
        Index of the `z` element in the recipe to use in the filename.

    Returns
    -------
    path : str
        Path of the file to load.

    """
    # check parameters
    check_parameter(recipe=dict,
                    input_folder=str,
                    fov=int,
                    r=int,
                    c=int,
                    z=int)

    # check if the recipe is fitted
    if not _is_recipe_fitted(recipe):
        recipe = fit_recipe(recipe)

    # build a map of the elements' indices
    map_element_index = {"fov": fov, "r": r, "c": c, "z": z}

    # get filename pattern and decompose it
    recipe_pattern = recipe["pattern"]
    path_elements = re.findall("fov|r|c|z|ext|opt", recipe_pattern)
    path_separators = re.split("fov|r|c|z|ext|opt", recipe_pattern)

    # get filename recombining elements of the recipe
    filename = path_separators[0]  # usually an empty string
    for (element_name, separator) in zip(path_elements, path_separators[1:]):

        # if we need an element from a list of elements of the same dimension
        # (eg. to pick a specific channel 'c' among a list of channels)
        if element_name in map_element_index:
            element_index = map_element_index[element_name]
            element = recipe[element_name][element_index]

        # if this element is unique for all the recipe (eg. 'fov')
        else:
            element = recipe[element_name]

        # the filename is built ensuring the order of apparition of the
        # different morphemes and their separators
        filename += element
        filename += separator

    # get path
    path = os.path.join(input_folder, filename)

    return path


def get_nb_element_per_dimension(recipe):
    """Count the number of element to stack for each dimension (`r`, `c`
    and `z`).

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Only contain the keys
        `fov`, `r`, `c`, `z`, `ext` or `opt`.

    Returns
    -------
    nb_r : int
        Number of rounds to be stacked.
    nb_c : int
        Number of channels to be stacked.
    nb_z : int
        Number of z layers to be stacked.

    """
    # check parameters
    check_parameter(recipe=dict)

    # check if the recipe is fitted
    if not _is_recipe_fitted(recipe):
        recipe = fit_recipe(recipe)

    return len(recipe["r"]), len(recipe["c"]), len(recipe["z"])


def count_nb_fov(recipe):
    """Count the number of different fields of view that can be defined from
    the recipe.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Can only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.

    Returns
    -------
    nb_fov : int
        Number of different fields of view in the recipe.

    """
    # check parameters
    check_parameter(recipe=dict)

    # check if the recipe is fitted
    if not _is_recipe_fitted(recipe):
        recipe = fit_recipe(recipe)

    # a good recipe should have a list in the 'fov' key
    if not isinstance(recipe["fov"], list):
        raise TypeError("'fov' should be a List or a str, not {0}"
                        .format(type(recipe["fov"])))
    else:
        return len(recipe["fov"])


def check_datamap(data_map):
    """Check and validate a data map.

    Checking a data map consists in validating the recipe-folder pairs.

    Parameters
    ----------
    data_map : List[tuple]
        Map between input directories and recipes.

    Returns
    -------
    _ : bool
        Assert if the data map is well formatted.

    """
    check_parameter(data_map=list)
    for pair in data_map:
        if not isinstance(pair, (tuple, list)):
            raise TypeError("A data map is a list with tuples or lists. "
                            "Not {0}".format(type(pair)))
        if len(pair) != 2:
            raise ValueError("Elements of a data map are tuples or lists that "
                             "map a recipe (dict) to an input directory "
                             "(string). Here {0} elements are given {1}"
                             .format(len(pair), pair))
        (recipe, input_folder) = pair
        if not isinstance(input_folder, str):
            raise TypeError("A data map map a recipe (dict) to an input "
                            "directory (string). Not ({0}, {1})"
                            .format(type(recipe), type(input_folder)))
        check_recipe(recipe, data_directory=input_folder)

    return True


# ### Sanity checks parameters ###

def check_parameter(**kwargs):
    """Check dtype of the function's parameters.

    Parameters
    ----------
    kwargs : Type or Tuple[Type]
        Map of each parameter with its expected dtype.

    Returns
    -------
    _ : bool
        Assert if the array is well formatted.

    """
    # get the frame and the parameters of the function
    frame = inspect.currentframe().f_back
    _, _, _, values = inspect.getargvalues(frame)

    # compare each parameter with its expected dtype
    for arg in kwargs:
        expected_dtype = kwargs[arg]
        parameter = values[arg]
        if not isinstance(parameter, expected_dtype):
            actual = "'{0}'".format(type(parameter).__name__)
            if isinstance(expected_dtype, tuple):
                target = ["'{0}'".format(x.__name__) for x in expected_dtype]
                target = "(" + ", ".join(target) + ")"
            else:
                target = expected_dtype.__name__
            raise TypeError("Parameter {0} should be a {1}. It is a {2} "
                            "instead.".format(arg, target, actual))

    return True


# ### Constants ###

def get_margin_value():
    """Return the margin pixel around a cell coordinate used to define its
    bounding box.

    Returns
    -------
    _ : int
        Margin value (in pixels).

    """
    # should be greater or equal to 2 (maybe 1 is enough)
    return 5


def get_eps_float32():
    """Return the epsilon value for a 32 bit float.

    Returns
    -------
    _ : np.float32
        Epsilon value.

    """
    return np.finfo(np.float32).eps


# ### Fetch data ###

def load_and_save_url(remote_url, directory, filename=None):
    """Download remote data and save them

    Parameters
    ----------
    remote_url : str
        Remote url of the data to download.
    directory : str
        Directory to save the download content.
    filename : str
        Filename of the object to save.

    Returns
    -------
    path : str
        Path of the downloaded file.

    """
    # check parameters
    check_parameter(remote_url=str,
                    directory=str,
                    filename=(str, type(None)))

    # get output path
    if filename is None:
        filename = remote_url.split("/")[-1]
    path = os.path.join(directory, filename)

    # download and save data
    urlretrieve(remote_url, path)

    return path


def check_hash(path, expected_hash):
    """Check hash value of a file.

    Parameters
    ----------
    path : str
        Path of the file to check.
    expected_hash : str
        Expected hash value.

    Returns
    -------
    _ : bool
        True if hash values match.

    """
    # check parameter
    check_parameter(path=str,
                    expected_hash=str)

    # compute hash value
    hash_value = compute_hash(path)

    # compare checksum
    if hash_value != expected_hash:
        raise IOError("File {0} has an SHA256 checksum ({1}) differing from "
                      "expected ({2}). File may be corrupted."
                      .format(path, hash_value, expected_hash))

    return True


def compute_hash(path):
    """Compute sha256 hash of a file.

    Parameters
    ----------
    path : str
        Path to read the file.

    Returns
    -------
    sha256 : str
        Hash value of the file.

    """
    # check parameters
    check_parameter(path=str)

    # initialization
    sha256hash = hashlib.sha256()
    chunk_size = 8192

    # open and read file
    with open(path, "rb") as f:
        while True:
            buffer = f.read(chunk_size)
            if not buffer:
                break
            sha256hash.update(buffer)

    # compute hash
    sha256 = sha256hash.hexdigest()

    return sha256


def check_input_data(input_directory, input_segmentation=False):
    """Check input images exists and download them if necessary.

    Parameters
    ----------
    input_directory : str
        Path of the image directory.
    input_segmentation : bool
        Check 2-d example images for segmentation.

    """
    # parameters
    filename_input_dapi = "experiment_1_dapi_fov_1.tif"
    url_input_dapi = "https://github.com/fish-quant/big-fish-examples/releases/download/data/experiment_1_dapi_fov_1.tif"
    hash_input_dapi = "3ce6dcfbece75da41326943432ada4cc9bacd06750e59dc2818bb253b6e7fdcd"
    filename_input_smfish = "experiment_1_smfish_fov_1.tif"
    url_input_smfish = "https://github.com/fish-quant/big-fish-examples/releases/download/data/experiment_1_smfish_fov_1.tif"
    hash_input_smfish = "bc6aec1f3da4c25f3c6b579c274584ce1e88112c7f980e5437b5ad5223bc8ff6"
    filename_input_nuc_full = "example_nuc_full.tif"
    url_input_nuc_full = "https://github.com/fish-quant/big-fish-examples/releases/download/data/example_nuc_full.tif"
    hash_input_nuc_full = "3bf70c7b5a02c60725baba3dfddff3010e0957de9ab78f0f65166248ead84ec4"
    filename_input_cell_full = "example_cell_full.tif"
    url_input_cell_full = "https://github.com/fish-quant/big-fish-examples/releases/download/data/example_cell_full.tif"
    hash_input_cell_full = "36981955ed97e9cab8a69241140a9aac3bdcf32dc157d6957fd37edcb16b34bd"

    # check if input dapi image exists
    path = os.path.join(input_directory, filename_input_dapi)
    if os.path.isfile(path):

        # check that image is not corrupted
        try:
            check_hash(path, hash_input_dapi)
            print("{0} is already in the directory"
                  .format(filename_input_dapi))

        # otherwise download it
        except IOError:
            print("{0} seems corrupted".format(filename_input_dapi))
            print("downloading {0}...".format(filename_input_dapi))
            load_and_save_url(url_input_dapi,
                              input_directory,
                              filename_input_dapi)
            check_hash(path, hash_input_dapi)

    # if file does not exist we directly download it
    else:
        print("downloading {0}...".format(filename_input_dapi))
        load_and_save_url(url_input_dapi,
                          input_directory,
                          filename_input_dapi)
        check_hash(path, hash_input_dapi)

    # check if input smfish image exists
    path = os.path.join(input_directory, filename_input_smfish)
    if os.path.isfile(path):

        # check that image is not corrupted
        try:
            check_hash(path, hash_input_smfish)
            print("{0} is already in the directory"
                  .format(filename_input_smfish))

        # otherwise download it
        except IOError:
            print("{0} seems corrupted".format(filename_input_smfish))
            print("downloading {0}...".format(filename_input_smfish))
            load_and_save_url(url_input_smfish,
                              input_directory,
                              filename_input_smfish)
            check_hash(path, hash_input_smfish)

    # if file does not exist we directly download it
    else:
        print("downloading {0}...".format(filename_input_smfish))
        load_and_save_url(url_input_smfish,
                          input_directory,
                          filename_input_smfish)
        check_hash(path, hash_input_smfish)

    # stop here or check segmentation examples
    if input_segmentation:

        # check if example nucleus exists
        path = os.path.join(input_directory, filename_input_nuc_full)
        if os.path.isfile(path):

            # check that image is not corrupted
            try:
                check_hash(path, hash_input_nuc_full)
                print("{0} is already in the directory"
                      .format(filename_input_nuc_full))

            # otherwise download it
            except IOError:
                print("{0} seems corrupted".format(filename_input_nuc_full))
                print("downloading {0}...".format(filename_input_nuc_full))
                load_and_save_url(url_input_nuc_full,
                                  input_directory,
                                  filename_input_nuc_full)
                check_hash(path, hash_input_nuc_full)

        # if file does not exist we directly download it
        else:
            print("downloading {0}...".format(filename_input_nuc_full))
            load_and_save_url(url_input_nuc_full,
                              input_directory,
                              filename_input_nuc_full)
            check_hash(path, hash_input_nuc_full)

        # check if example cell exists
        path = os.path.join(input_directory, filename_input_cell_full)
        if os.path.isfile(path):

            # check that image is not corrupted
            try:
                check_hash(path, hash_input_cell_full)
                print("{0} is already in the directory"
                      .format(filename_input_cell_full))

            # otherwise download it
            except IOError:
                print("{0} seems corrupted".format(filename_input_cell_full))
                print("downloading {0}...".format(filename_input_cell_full))
                load_and_save_url(url_input_cell_full,
                                  input_directory,
                                  filename_input_cell_full)
                check_hash(path, hash_input_cell_full)

        # if file does not exist we directly download it
        else:
            print("downloading {0}...".format(filename_input_cell_full))
            load_and_save_url(url_input_cell_full,
                              input_directory,
                              filename_input_cell_full)
            check_hash(path, hash_input_cell_full)


# ### Computation ###

def moving_average(array, n):
    """Compute a trailing average.

    Parameters
    ----------
    array : np.ndarray
        Array used to compute moving average.
    n : int
        Window width of the moving average.

    Returns
    -------
    results : np.ndarray
        Moving average values.

    """
    # check parameter
    check_parameter(n=int)
    check_array(array, ndim=1)

    # compute moving average
    cumsum = [0]
    results = []
    for i, x in enumerate(array, 1):
        cumsum.append(cumsum[i-1] + x)
        if i >= n:
            ma = (cumsum[i] - cumsum[i - n]) / n
            results.append(ma)
    results = np.array(results)

    return results


def centered_moving_average(array, n):
    """Compute a centered moving average.

    Parameters
    ----------
    array : np.ndarray
        Array used to compute moving average.
    n : int
        Window width of the moving average.

    Returns
    -------
    results : np.ndarray
        Centered moving average values.

    """
    # check parameter
    check_parameter(n=int)
    check_array(array, ndim=1)

    # pad array to keep the same length and centered the outcome
    if n % 2 == 0:
        r = int(n / 2)
        n += 1
    else:
        r = int((n - 1) / 2)
    array_padded = np.pad(array, pad_width=r, mode="reflect")

    # compute centered moving average
    results = moving_average(array_padded, n)

    return results


# ### Spot utilities ###

def get_sigma(voxel_size_z=None, voxel_size_yx=100, psf_z=None, psf_yx=200):
    """Compute the standard deviation of the PSF of the spots.

    Parameters
    ----------
    voxel_size_z : int or float or None
        Height of a voxel, along the z axis, in nanometer. If None, we consider
        a 2-d PSF.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    psf_z : int or float or None
        Theoretical size of the PSF emitted by a spot in the z plan, in
        nanometer. If None, we consider a 2-d PSF.
    psf_yx : int or float
        Theoretical size of the PSF emitted by a spot in the yx plan, in
        nanometer.

    Returns
    -------
    sigma : Tuple[float]
        Standard deviations in pixel of the PSF, one element per dimension.

    """
    # check parameters
    check_parameter(voxel_size_z=(int, float, type(None)),
                    voxel_size_yx=(int, float),
                    psf_z=(int, float, type(None)),
                    psf_yx=(int, float))

    # compute sigma
    sigma_yx = psf_yx / voxel_size_yx

    if voxel_size_z is None or psf_z is None:
        return sigma_yx, sigma_yx

    else:
        sigma_z = psf_z / voxel_size_z
        return sigma_z, sigma_yx, sigma_yx


def get_radius(voxel_size_z=None, voxel_size_yx=100, psf_z=None, psf_yx=200):
    """Approximate the radius of the detected spot.

    We use the formula:

    .. math::

        \\mbox{radius} = \\mbox{sqrt(ndim)} * \\sigma

    with :math:`\\mbox{ndim}` the number of dimension of the image and
    :math:`\\sigma` the standard deviation (in pixel) of the detected spot.

    Parameters
    ----------
    voxel_size_z : int or float or None
        Height of a voxel, along the z axis, in nanometer. If None, we consider
        a 2-d spot.
    voxel_size_yx : int or float
        Size of a voxel on the yx plan, in nanometer.
    psf_z : int or float or None
        Theoretical size of the PSF emitted by a spot in the z plan, in
        nanometer. If None, we consider a 2-d spot.
    psf_yx : int or float
        Theoretical size of the PSF emitted by a spot in the yx plan, in
        nanometer.

    Returns
    -------
    radius : Tuple[float]
        Radius in pixels of the detected spots, one element per dimension.

    """
    # compute sigma
    sigma = get_sigma(voxel_size_z, voxel_size_yx, psf_z, psf_yx)

    # compute radius
    radius = [np.sqrt(len(sigma)) * sigma_ for sigma_ in sigma]
    radius = tuple(radius)

    return radius



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.detection.spot_modeling module.
"""

# TODO test bigfish.detection.fit_subpixel
# TODO test bigfish.detection.build_reference_spot
# TODO test bigfish.detection.modelize_spot
# TODO test bigfish.detection.precompute_erf



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions used to format and clean any intermediate results loaded in or
returned by a bigfish method.
"""

import warnings

import numpy as np
import pandas as pd
from scipy import ndimage as ndi

from .utils import check_array, check_parameter, get_margin_value
from .io import save_data_to_csv

from skimage.measure import regionprops, find_contours
from skimage.draw import polygon_perimeter


# ### Object identification to sub-cellular regions ###

def identify_objects_in_region(mask, coord, ndim):
    """Identify cellular objects in specific region.

    Parameters
    ----------
    mask : np.ndarray, bool
        Binary mask of the targeted region with shape (y, x).
    coord : np.ndarray
        Array with two dimensions. One object per row, zyx or yx coordinates
        in the first 3 or 2 columns.
    ndim : int
        Number of spatial dimensions to consider (2 or 3).

    Returns
    -------
    coord_in : np.ndarray
        Coordinates of the objects detected inside the region.
    coord_out : np.ndarray
        Coordinates of the objects detected outside the region.

    """
    # check parameters
    check_parameter(ndim=int)
    check_array(mask,
                ndim=2,
                dtype=[np.uint8, np.uint16, np.int64,
                       bool])
    check_array(coord,
                ndim=2,
                dtype=[np.int64, np.float64])

    # check number of dimensions
    if ndim not in [2, 3]:
        raise ValueError("The number of spatial dimension requested should be "
                         "2 or 3, not {0}.".format(ndim))
    if coord.shape[1] < ndim:
        raise ValueError("Coord array should have at least {0} features to "
                         "match the number of spatial dimensions requested. "
                         "Currently {1} is not enough."
                         .format(ndim, coord.shape[1]))

    # binarize nuclei mask if needed
    if mask.dtype != bool:
        mask = mask.astype(bool)

    # cast coordinates dtype if necessary
    if coord.dtype == np.int64:
        coord_int = coord
    else:
        coord_int = np.round(coord).astype(np.int64)

    # remove objects inside the region
    mask_in = mask[coord_int[:, ndim - 2], coord_int[:, ndim - 1]]
    coord_in = coord[mask_in]
    coord_out = coord[~mask_in]

    return coord_in, coord_out


def remove_transcription_site(rna, clusters, nuc_mask, ndim):
    """Distinguish RNA molecules detected in a transcription site from the
    rest.

    A transcription site is defined as as a foci detected within the nucleus.

    Parameters
    ----------
    rna : np.ndarray
        Coordinates of the detected RNAs with shape (nb_spots, 4) or
        (nb_spots, 3). One coordinate per dimension (zyx or yx coordinates)
        plus the index of the cluster assigned to the RNA. If no cluster was
        assigned, value is -1.
    clusters : np.ndarray
        Array with shape (nb_clusters, 5) or (nb_clusters, 4). One coordinate
        per dimension for the clusters centroid (zyx or yx coordinates),
        the number of RNAs detected in the clusters and their index.
    nuc_mask : np.ndarray, bool
        Binary mask of the nuclei region with shape (y, x).
    ndim : int
        Number of spatial dimensions to consider (2 or 3).

    Returns
    -------
    rna_out_ts : np.ndarray
        Coordinates of the detected RNAs with shape (nb_spots, 4) or
        (nb_spots, 3). One coordinate per dimension (zyx or yx coordinates)
        plus the index of the foci assigned to the RNA. If no foci was
        assigned, value is -1. RNAs from transcription sites are removed.
    foci : np.ndarray
        Array with shape (nb_foci, 5) or (nb_foci, 4). One coordinate per
        dimension for the foci centroid (zyx or yx coordinates),
        the number of RNAs detected in the foci and its index.
    ts : np.ndarray
        Array with shape (nb_ts, 5) or (nb_ts, 4). One coordinate per
        dimension for the transcription site centroid (zyx or yx coordinates),
        the number of RNAs detected in the transcription site and its index.

    """
    # check parameters
    check_array(rna,
                ndim=2,
                dtype=[np.int64, np.float64])

    # discriminate foci from transcription sites
    ts, foci = identify_objects_in_region(
        nuc_mask, clusters, ndim)

    # filter out rna from transcription sites
    rna_in_ts = ts[:, ndim + 1]
    mask_rna_in_ts = np.isin(rna[:, ndim], rna_in_ts)
    rna_out_ts = rna[~mask_rna_in_ts]

    return rna_out_ts, foci, ts


# ### Cell extraction ###

def extract_cell(cell_label, ndim, nuc_label=None, rna_coord=None,
                 others_coord=None, image=None, others_image=None,
                 remove_cropped_cell=True, check_nuc_in_cell=True):
    """Extract cell-level results for an image.

    The function gathers different segmentation and detection results obtained
    at the image level and assigns each of them to the individual cells.

    Parameters
    ----------
    cell_label : np.ndarray, np.uint or np.int
        Image with labelled cells and shape (y, x).
    ndim : int
        Number of spatial dimensions to consider (2 or 3).
    nuc_label : np.ndarray, np.uint or np.int
        Image with labelled nuclei and shape (y, x). If None, individual
        nuclei are not assigned to each cell.
    rna_coord : np.ndarray
        Coordinates of the detected RNAs with zyx or yx coordinates in the
        first 3 or 2 columns. If None, RNAs are not assigned to individual
        cells.
    others_coord : Dict[np.ndarray]
        Dictionary of coordinates arrays. For each array of the dictionary,
        the different elements are assigned to individual cells. Arrays should
        be organized the same way than spots: zyx or yx coordinates in the
        first 3 or 2 columns, np.int64 dtype, one element per row. Can be used
        to assign different detected elements to the segmented cells along with
        the spots. If None, no others elements are assigned to the individual
        cells.
    image : np.ndarray, np.uint
        Image in 2-d. If None, image of the individual cells are not extracted.
    others_image : Dict[np.ndarray]
        Dictionary of images to crop. If None, no others image of the
        individual cells are extracted.
    remove_cropped_cell : bool
        Remove cells cropped by the FoV frame.
    check_nuc_in_cell : bool
        Check that each nucleus is entirely localized within a cell.

    Returns
    -------
    fov_results : List[Dict]
        List of dictionaries, one per cell segmented in the image. Each
        dictionary includes information about the cell (image, masks,
        coordinates arrays). Minimal information are:

        * `cell_id`: Unique id of the cell.
        * `bbox`: bounding box coordinates with the order (`min_y`, `min_x`,
          `max_y`, `max_x`).
        * `cell_coord`: boundary coordinates of the cell.
        * `cell_mask`: mask of the cell.

    """
    # check parameters
    check_parameter(ndim=int,
                    others_coord=(dict, type(None)),
                    others_image=(dict, type(None)),
                    remove_cropped_cell=bool,
                    check_nuc_in_cell=bool)
    check_array(cell_label, ndim=2, dtype=[np.uint8, np.uint16, np.int64])
    if nuc_label is not None:
        check_array(nuc_label, ndim=2, dtype=[np.uint8, np.uint16, np.int64])
    if rna_coord is not None:
        check_array(rna_coord, ndim=2, dtype=[np.int64, np.float64])
    if image is not None:
        check_array(image, ndim=2, dtype=[np.uint8, np.uint16])
    actual_keys = ["cell_id", "bbox", "cell_coord", "cell_mask", "nuc_coord",
                   "nuc_mask", "rna_coord", "image"]
    if others_coord is not None:
        for key in others_coord:
            if key in actual_keys:
                raise KeyError("Key {0} in 'others_coord' is already taken. "
                               "Please choose another one.".format(key))
            else:
                actual_keys.append(key)
            array = others_coord[key]
            check_array(array, ndim=2, dtype=[np.int64, np.float64])
            if array.shape[1] < ndim:
                warnings.warn("Array in 'others_coord' have less coordinates "
                              "({0}) than the minimum number of spatial "
                              "dimension we consider ({1})."
                              .format(array.shape[1], ndim),
                              UserWarning)
    # TODO allow boolean for 'others_image'
    # TODO bug if 'image' is None but not 'others_image'
    if others_image is not None:
        for key in others_image:
            if key in actual_keys:
                raise KeyError("Key {0} in 'others_image' is already taken. "
                               "Please choose another one.".format(key))
            else:
                actual_keys.append(key)
            image_ = others_image[key]
            check_array(image_, ndim=2, dtype=[np.uint8, np.uint16])
            if image_.shape != image.shape:
                warnings.warn("Image in 'others_image' does not have the same "
                              "shape ({0}) than original image ({1})."
                              .format(image_.shape, image.shape),
                              UserWarning)
    if rna_coord.shape[1] < ndim:
        warnings.warn("'rna_coord' have less coordinates ({0}) than the "
                      "minimum number of spatial dimension we "
                      "consider ({1}).".format(rna_coord.shape[1], ndim),
                      UserWarning)

    # initialize FoV results
    fov_results = []

    # initialize a mask to detect cells at the FoV borders
    fov_borders = np.zeros(cell_label.shape, dtype=bool)
    if remove_cropped_cell:
        fov_borders[:, 0] = True
        fov_borders[0, :] = True
        fov_borders[:, cell_label.shape[1] - 1] = True
        fov_borders[cell_label.shape[0] - 1, :] = True

    # iterate over each segmented cell
    cells = regionprops(cell_label)
    for cell in cells:

        # initialize cell results
        cell_results = {}

        # get the bounding box of the cell
        label = cell.label
        cell_results["cell_id"] = label
        (min_y, min_x, max_y, max_x) = cell.bbox
        cell_results["bbox"] = cell.bbox

        # get binary masks of the cell
        cell_mask = cell_label.copy()
        cell_mask = (cell_mask == label)

        # check if cell is not cropped by the borders
        if remove_cropped_cell and _check_cropped_cell(cell_mask, fov_borders):
            continue

        # get boundaries coordinates for cell
        cell_coord = from_binary_to_coord(cell_mask)
        cell_coord = complete_coord_boundaries(cell_coord)
        cell_coord[:, 0] -= min_y
        cell_coord[:, 1] -= min_x
        cell_results["cell_coord"] = cell_coord

        # crop binary mask of the cell
        cell_mask_cropped = cell_mask[min_y: max_y, min_x: max_x]
        cell_results["cell_mask"] = cell_mask_cropped

        # get binary mask of the nucleus
        if nuc_label is not None:
            nuc_mask = nuc_label.copy()
            nuc_mask = (nuc_mask == label)

            # check if nucleus is in the cell
            if (check_nuc_in_cell
                    and not _check_nucleus_in_cell(cell_mask, nuc_mask)):
                continue

            # get boundaries coordinates for nucleus
            nuc_coord = from_binary_to_coord(nuc_mask)
            nuc_coord = complete_coord_boundaries(nuc_coord)
            nuc_coord[:, 0] -= min_y
            nuc_coord[:, 1] -= min_x
            cell_results["nuc_coord"] = nuc_coord

            # crop binary mask of the nucleus
            nuc_mask_cropped = nuc_mask[min_y: max_y, min_x: max_x]
            cell_results["nuc_mask"] = nuc_mask_cropped

        # get coordinates of the spots detected in the cell
        if rna_coord is not None:
            rna_in_cell, _ = identify_objects_in_region(
                cell_mask, rna_coord, ndim)
            rna_in_cell[:, ndim - 2] -= min_y
            rna_in_cell[:, ndim - 1] -= min_x
            cell_results["rna_coord"] = rna_in_cell

        # get coordinates of the other detected elements
        if others_coord is not None:
            for key in others_coord:
                array = others_coord[key]
                element_in_cell, _ = identify_objects_in_region(
                    cell_mask, array, ndim)
                element_in_cell[:, ndim - 2] -= min_y
                element_in_cell[:, ndim - 1] -= min_x
                cell_results[key] = element_in_cell

        # crop cell image
        if image is not None:
            image_cropped = image[min_y: max_y, min_x: max_x]
            cell_results["image"] = image_cropped

        # get crops of the other images
        if others_image is not None:
            for key in others_image:
                image_ = others_image[key]
                image_cropped_ = image_[min_y: max_y, min_x: max_x]
                cell_results[key] = image_cropped_

        fov_results.append(cell_results)

    return fov_results


def _check_cropped_cell(cell_mask, border_frame):
    """
    Check if a cell is cropped by the border frame.

    Parameters
    ----------
    cell_mask : np.ndarray, bool
        Binary mask of the cell cytoplasm.

    border_frame : np.ndarray, bool
        Binary mask of the border frame.

    Returns
    -------
    _ : bool
        True if cell is cropped.

    """
    # check cell is not cropped by the borders
    crop = cell_mask & border_frame
    if np.any(crop):
        return True
    else:
        return False


def _check_nucleus_in_cell(cell_mask, nuc_mask):
    """
    Check if the nucleus is properly contained in the cell cytoplasm.

    Parameters
    ----------
    cell_mask : np.ndarray, bool
        Binary mask of the cell cytoplasm.

    nuc_mask : np.ndarray, bool
        Binary mask of the nucleus cytoplasm.

    Returns
    -------
    _ : bool
        True if the nucleus is in the cell.

    """
    diff = cell_mask | nuc_mask
    if np.any(diff != cell_mask):
        return False
    else:
        return True


def extract_spots_from_frame(spots, z_lim=None, y_lim=None, x_lim=None):
    """Get spots coordinates within a given frame.

    Parameters
    ----------
    spots : np.ndarray
        Coordinate of the spots detected inside foci, with shape (nb_spots, 3)
        or (nb_spots, 4). One coordinate per dimension (zyx coordinates) plus
        the index of the foci if necessary.
    z_lim : tuple[int, int]
        Minimum and maximum coordinate of the frame along the z axis.
    y_lim : tuple[int, int]
        Minimum and maximum coordinate of the frame along the y axis.
    x_lim : tuple[int, int]
        Minimum and maximum coordinate of the frame along the x axis.

    Returns
    -------
    extracted_spots : np.ndarray
        Coordinate of the spots detected inside foci, with shape (nb_spots, 3)
        or (nb_spots, 4). One coordinate per dimension (zyx coordinates) plus
        the index of the foci if necessary.

    """
    # check parameters
    check_array(spots,
                ndim=2,
                dtype=[np.int64, np.float64])
    check_parameter(z_lim=(tuple, type(None)),
                    y_lim=(tuple, type(None)),
                    x_lim=(tuple, type(None)))

    # extract spots
    extracted_spots = spots.copy()
    if z_lim is not None:
        extracted_spots = extracted_spots[extracted_spots[:, 0] < z_lim[1]]
        extracted_spots = extracted_spots[z_lim[0] < extracted_spots[:, 0]]
        extracted_spots[:, 0] -= z_lim[0]
    if y_lim is not None:
        extracted_spots = extracted_spots[extracted_spots[:, 1] < y_lim[1]]
        extracted_spots = extracted_spots[y_lim[0] < extracted_spots[:, 1]]
        extracted_spots[:, 1] -= y_lim[0]
    if x_lim is not None:
        extracted_spots = extracted_spots[extracted_spots[:, 2] < x_lim[1]]
        extracted_spots = extracted_spots[x_lim[0] < extracted_spots[:, 2]]
        extracted_spots[:, 2] -= x_lim[0]

    return extracted_spots


def summarize_extraction_results(fov_results, ndim, path_output=None):
    """Summarize results extracted from an image and store them in a dataframe.

    Parameters
    ----------
    fov_results : List[Dict]
        List of dictionaries, one per cell segmented in the image. Each
        dictionary includes information about the cell (image, masks,
        coordinates arrays). Minimal information are:

        * `cell_id`: Unique id of the cell.
        * `bbox`: bounding box coordinates with the order (`min_y`, `min_x`,
          `max_y`, `max_x`).
        * `cell_coord`: boundary coordinates of the cell.
        * `cell_mask`: mask of the cell.
    ndim : int
        Number of spatial dimensions to consider (2 or 3).
    path_output : str or None
        Path to save the dataframe in a csv file.

    Returns
    -------
    df : pd.DataFrame
        Dataframe with summarized results from the field of view, at the cell
        level. At least `cell_id` (Unique id of the cell) is returned. Other
        indicators are summarized if available:

        * `nb_rna`: Number of detected rna in the cell.
        * `nb_rna_in_nuc`: Number of detected rna inside the nucleus.
        * `nb_rna_out_nuc`: Number of detected rna outside the nucleus.

        Extra coordinates elements detected are counted in the cell and
        summarized as well.

    """
    # check parameters
    check_parameter(fov_results=list,
                    ndim=int,
                    path_output=(str, type(None)))

    # case if no cell were detected
    # TODO make it consistent with the case where there are cells
    if len(fov_results) == 0:
        df = pd.DataFrame({"cell_id": []})
        if path_output is not None:
            save_data_to_csv(df, path_output)
        return df

    # check extra coordinates to summarize
    cell_results = fov_results[0]
    _extra_coord = {}
    for key in cell_results:
        if key in ["cell_id", "bbox", "cell_coord", "cell_mask",
                   "nuc_coord", "nuc_mask", "rna_coord", "image"]:
            continue
        others_coord = cell_results[key]
        if (not isinstance(others_coord, np.ndarray)
                or others_coord.dtype not in [np.int64, np.float64]):
            continue
        _extra_coord[key] = []

    # summarize results at the cell level
    _cell_id = []
    _nb_rna = []
    _nb_rna_in_nuc = []
    _nb_rna_out_nuc = []
    for cell_results in fov_results:
        # get cell id
        _cell_id.append(cell_results["cell_id"])

        # get rna coordinates and relative results
        if "rna_coord" in cell_results:
            rna_coord = cell_results["rna_coord"]
            _nb_rna.append(len(rna_coord))

            # get rna in nucleus
            if "nuc_mask" in cell_results:
                nuc_mask = cell_results["nuc_mask"]
                rna_in_nuc, rna_out_nuc = identify_objects_in_region(
                    nuc_mask, rna_coord, ndim)
                _nb_rna_in_nuc.append(len(rna_in_nuc))
                _nb_rna_out_nuc.append(len(rna_out_nuc))

        # get others coordinates
        for key in _extra_coord:
            others_coord = cell_results[key]
            _extra_coord[key].append(len(others_coord))

    # complete missing mandatory results
    n = len(_cell_id)
    if len(_nb_rna) == 0:
        _nb_rna = [np.nan] * n
    if len(_nb_rna_in_nuc) == 0:
        _nb_rna_in_nuc = [np.nan] * n
    if len(_nb_rna_out_nuc) == 0:
        _nb_rna_out_nuc = [np.nan] * n

    # store minimum results in a dataframe
    result_summary = {"cell_id": _cell_id,
                      "nb_rna": _nb_rna,
                      "nb_rna_in_nuc": _nb_rna_in_nuc,
                      "nb_rna_out_nuc": _nb_rna_out_nuc}

    # store available results on nucleus and rna
    if len(_nb_rna) > 0:
        result_summary["nb_rna"] = _nb_rna
    if len(_nb_rna_in_nuc) > 0:
        result_summary["nb_rna_in_nuc"] = _nb_rna_in_nuc
    if len(_nb_rna_out_nuc) > 0:
        result_summary["nb_rna_out_nuc"] = _nb_rna_out_nuc

    # store results from others elements detected in the cell
    for key in _extra_coord:
        result_summary["nb_{0}".format(key)] = _extra_coord[key]

    # instantiate dataframe
    df = pd.DataFrame(result_summary)

    # save dataframe
    if path_output is not None:
        save_data_to_csv(df, path_output)

    return df


# ### Segmentation postprocessing ###

def center_mask_coord(main, others=None):
    """Center a 2-d binary mask (surface or boundaries) or a 2-d localization
    coordinates array and pad it.

    One mask or coordinates array should be at least provided (`main`). If
    others masks or arrays are provided (`others`), they will be transformed
    like `main`. All the provided masks should have the same shape.

    Parameters
    ----------
    main : np.ndarray, np.uint or np.int or bool
        Binary image with shape (y, x) or array of coordinates with shape
        (nb_points, 2).
    others : List(np.ndarray)
        List of binary image with shape (y, x), array of coordinates with
        shape (nb_points, 2) or array of coordinates with shape (nb_points, 3).

    Returns
    -------
    main_centered : np.ndarray, np.uint or np.int or bool
        Centered binary image with shape (y, x).
    others_centered : List(np.ndarray)
        List of centered binary image with shape (y, x), centered array of
        coordinates with shape (nb_points, 2) or centered array of coordinates
        with shape (nb_points, 3).

    """
    # TODO allow the case when coordinates do not represent external boundaries
    # check parameters
    check_array(main,
                ndim=2,
                dtype=[np.uint8, np.uint16, np.int64, bool])
    check_parameter(others=(list, type(None)))
    if others is not None and len(others) != 0:
        for x in others:
            if x is None:
                continue
            check_array(x,
                        ndim=2,
                        dtype=[np.uint8, np.uint16, np.int64, bool])

    # initialize parameter
    marge = get_margin_value()

    # compute by how much we need to move the main object to center it
    if main.shape[1] == 2:
        # 'main' is already a 2-d coordinates array (probably locating the
        # external boundaries of the main object)
        main_coord = main.copy()
    else:
        # get external boundaries coordinates
        main_coord = from_binary_to_coord(main)

    # we get the minimum and maximum from external boundaries coordinates so
    # we add 1 to the minimum and substract 1 to the maximum
    min_y, max_y = main_coord[:, 0].min() + 1, main_coord[:, 0].max() - 1
    min_x, max_x = main_coord[:, 1].min() + 1, main_coord[:, 1].max() - 1

    # we compute the shape of the main object with a predefined marge
    shape_y = max_y - min_y + 1
    shape_x = max_x - min_x + 1
    main_centered_shape = (shape_y + 2 * marge, shape_x + 2 * marge)

    # center the main object
    if main.shape[1] == 2:
        # 'main' is a 2-d coordinates array
        main_centered = main.copy()
        main_centered[:, 0] = main_centered[:, 0] - min_y + marge
        main_centered[:, 1] = main_centered[:, 1] - min_x + marge
    else:
        # 'main' is a 2-d binary matrix
        main_centered = np.zeros(main_centered_shape, dtype=main.dtype)
        crop = main[min_y:max_y + 1, min_x:max_x + 1]
        main_centered[marge:shape_y + marge, marge:shape_x + marge] = crop

    if others is None or len(others) == 0:
        return main_centered, None

    # center the others objects with the same transformation
    others_centered = []
    for other in others:
        if other is None:
            other_centered = None
        elif other.shape[1] == 2:
            # 'other' is a 2-d coordinates array
            other_centered = other.copy()
            other_centered[:, 0] = other_centered[:, 0] - min_y + marge
            other_centered[:, 1] = other_centered[:, 1] - min_x + marge
        elif other.shape[1] == 3 or other.shape[1] == 4:
            # 'other' is a 3-d or 4-d coordinates
            other_centered = other.copy()
            other_centered[:, 1] = other_centered[:, 1] - min_y + marge
            other_centered[:, 2] = other_centered[:, 2] - min_x + marge
        else:
            # 'other' is a 2-d binary matrix
            other_centered = np.zeros(main_centered_shape, dtype=other.dtype)
            crop = other[min_y:max_y + 1, min_x:max_x + 1]
            other_centered[marge:shape_y + marge, marge:shape_x + marge] = crop
        others_centered.append(other_centered)

    return main_centered, others_centered


def from_boundaries_to_surface(binary_boundaries):
    """Fill in the binary matrix representing the boundaries of an object.

    Parameters
    ----------
    binary_boundaries : np.ndarray, np.uint or np.int or bool
        Binary image with shape (y, x).

    Returns
    -------
    binary_surface : np.ndarray, bool
        Binary image with shape (y, x).

    """
    # check parameters
    check_array(binary_boundaries,
                ndim=2,
                dtype=[np.uint8, np.uint16, np.int64, bool])

    # from binary boundaries to binary surface
    binary_surface = ndi.binary_fill_holes(binary_boundaries)

    return binary_surface


def from_surface_to_boundaries(binary_surface):
    """Convert the binary surface to binary boundaries.

    Parameters
    ----------
    binary_surface : np.ndarray, np.uint or np.int or bool
        Binary image with shape (y, x).

    Returns
    -------
    binary_boundaries : np.ndarray, np.uint or np.int or bool
        Binary image with shape (y, x).

    """
    # check parameters
    check_array(binary_surface,
                ndim=2,
                dtype=[np.uint8, np.uint16, np.int64, bool])
    original_dtype = binary_surface.dtype

    # pad the binary surface in case object if on the edge
    binary_surface_ = np.pad(binary_surface, [(1, 1)], mode="constant")

    # compute distance map of the surface binary mask
    distance_map = ndi.distance_transform_edt(binary_surface_)

    # get binary boundaries
    binary_boundaries_ = (distance_map < 2) & (distance_map > 0)
    binary_boundaries_ = binary_boundaries_.astype(original_dtype)

    # remove pad
    binary_boundaries = binary_boundaries_[1:-1, 1:-1]

    return binary_boundaries


def from_binary_to_coord(binary):
    """Extract coordinates from a 2-d binary matrix.

    As the resulting coordinates represent the external boundaries of the
    object, the coordinates values can be negative.

    Parameters
    ----------
    binary : np.ndarray, np.uint or np.int or bool
        Binary image with shape (y, x).

    Returns
    -------
    coord : np.ndarray, np.int64
        Array of boundaries coordinates with shape (nb_points, 2).

    """
    # check parameters
    check_array(binary,
                ndim=2,
                dtype=[np.uint8, np.uint16, np.int64, bool])

    # we enlarge the binary mask with one pixel to be sure the external
    # boundaries of the object still fit within the frame
    binary_ = np.pad(binary, [(1, 1)], mode="constant")

    # get external boundaries coordinates
    coord = find_contours(binary_, level=0)[0].astype(np.int64)

    # remove the pad
    coord -= 1

    return coord


def complete_coord_boundaries(coord):
    """Complete a 2-d coordinates array, by generating/interpolating missing
    points.

    Parameters
    ----------
    coord : np.ndarray, np.int64
        Array of coordinates to complete, with shape (nb_points, 2).

    Returns
    -------
    coord_completed : np.ndarray, np.int64
        Completed coordinates arrays, with shape (nb_points, 2).

    """
    # check parameters
    check_array(coord,
                ndim=2,
                dtype=[np.int64])

    # for each array in the list, complete its coordinates using the scikit
    # image method 'polygon_perimeter'
    coord_y, coord_x = polygon_perimeter(coord[:, 0], coord[:, 1])
    coord_y = coord_y[:, np.newaxis]
    coord_x = coord_x[:, np.newaxis]
    coord_completed = np.concatenate((coord_y, coord_x), axis=-1)

    return coord_completed


def from_coord_to_frame(coord, external_coord=True):
    """Initialize a frame shape to represent coordinates values in 2-d matrix.

    If coordinates represent the external boundaries of an object, we add 1 to
    the minimum coordinate and substract 1 to the maximum coordinate in order
    to build the frame. The frame centers the coordinates by default.

    Parameters
    ----------
    coord : np.ndarray, np.int64
        Array of cell boundaries coordinates with shape (nb_points, 2) or
        (nb_points, 3).
    external_coord : bool
        Coordinates represent external boundaries of object.

    Returns
    -------
    frame_shape : tuple
        Shape of the 2-d matrix.
    min_y : int
        Value tu substract from the y coordinate axis.
    min_x : int
        Value tu substract from the x coordinate axis.
    marge : int
        Value to add to the coordinates.

    """
    # check parameter
    check_parameter(external_coord=bool)

    # initialize marge
    marge = get_margin_value()

    # from 2D coordinates boundaries to binary boundaries
    if external_coord:
        min_y, max_y = coord[:, 0].min() + 1, coord[:, 0].max() - 1
        min_x, max_x = coord[:, 1].min() + 1, coord[:, 1].max() - 1
    else:
        min_y, max_y = coord[:, 0].min(), coord[:, 0].max()
        min_x, max_x = coord[:, 1].min(), coord[:, 1].max()
    shape_y = max_y - min_y + 1
    shape_x = max_x - min_x + 1
    frame_shape = (shape_y + 2 * marge, shape_x + 2 * marge)

    return frame_shape, min_y, min_x, marge


def from_coord_to_surface(cyt_coord, nuc_coord=None, rna_coord=None,
                          external_coord=True):
    """Convert 2-d coordinates to a binary matrix with the surface of the
    object.

    If we manipulate the coordinates of the external boundaries, the relative
    binary matrix has two extra pixels in each dimension. We compensate by
    keeping only the inside pixels of the object surface.

    If others coordinates are provided (nucleus and mRNAs), the relative
    binary matrix is built with the same shape as the main coordinates (cell).

    Parameters
    ----------
    cyt_coord : np.ndarray, np.int64
        Array of cytoplasm boundaries coordinates with shape (nb_points, 2).
    nuc_coord : np.ndarray, np.int64
        Array of nucleus boundaries coordinates with shape (nb_points, 2).
    rna_coord : np.ndarray, np.int64
        Array of mRNAs coordinates with shape (nb_points, 2) or
        (nb_points, 3).
    external_coord : bool
        Coordinates represent external boundaries of object.

    Returns
    -------
    cyt_surface : np.ndarray, bool
        Binary image of cytoplasm surface with shape (y, x).
    nuc_surface : np.ndarray, bool
        Binary image of nucleus surface with shape (y, x).
    rna_binary : np.ndarray, bool
        Binary image of mRNAs localizations with shape (y, x).
    new_rna_coord : np.ndarray, np.int64
        Array of mRNAs coordinates with shape (nb_points, 2) or (nb_points, 3).

    """
    # check parameters
    check_array(cyt_coord,
                ndim=2,
                dtype=[np.int64])
    if nuc_coord is not None:
        check_array(nuc_coord,
                    ndim=2,
                    dtype=[np.int64])
    if rna_coord is not None:
        check_array(rna_coord,
                    ndim=2,
                    dtype=[np.int64])
    check_parameter(external_coord=bool)

    # center coordinates
    cyt_coord_, [nuc_coord_, rna_coord_] = center_mask_coord(
        main=cyt_coord,
        others=[nuc_coord, rna_coord])

    # get the binary frame
    frame_shape, min_y, min_x, marge = from_coord_to_frame(
        coord=cyt_coord_,
        external_coord=external_coord)

    # from coordinates to binary external boundaries
    cyt_boundaries_ext = np.zeros(frame_shape, dtype=bool)
    cyt_boundaries_ext[cyt_coord_[:, 0], cyt_coord_[:, 1]] = True
    if nuc_coord_ is not None:
        nuc_boundaries_ext = np.zeros(frame_shape, dtype=bool)
        nuc_boundaries_ext[nuc_coord_[:, 0], nuc_coord_[:, 1]] = True
    else:
        nuc_boundaries_ext = None

    # from binary external boundaries to binary external surface
    cyt_surface_ext = from_boundaries_to_surface(cyt_boundaries_ext)
    if nuc_boundaries_ext is not None:
        nuc_surface_ext = from_boundaries_to_surface(nuc_boundaries_ext)
    else:
        nuc_surface_ext = None

    # from binary external surface to binary surface
    cyt_surface = cyt_surface_ext & (~cyt_boundaries_ext)
    if nuc_surface_ext is not None:
        nuc_surface = nuc_surface_ext & (~nuc_boundaries_ext)
    else:
        nuc_surface = None

    # center mRNAs coordinates
    if rna_coord_ is not None:
        rna_binary = np.zeros(frame_shape, dtype=bool)
        if rna_coord_.shape[1] == 2:
            rna_binary[rna_coord_[:, 0], rna_coord_[:, 1]] = True
        else:
            rna_binary[rna_coord_[:, 1], rna_coord_[:, 2]] = True
        new_rna_coord = rna_coord_.copy()

    else:
        rna_binary = None
        new_rna_coord = None

    return cyt_surface, nuc_surface, rna_binary, new_rna_coord
